"""
Module for managing Windows systems and getting Windows system information.
Support for reboot, shutdown, join domain, rename

:depends:
    - pywintypes
    - win32api
    - win32con
    - win32net
    - wmi
"""

import ctypes
import logging
import platform
import time
from datetime import datetime

import salt.utils.functools
import salt.utils.locales
import salt.utils.platform
import salt.utils.win_system
import salt.utils.winapi
from salt.exceptions import CommandExecutionError

try:
    from ctypes import windll

    import pywintypes
    import win32api
    import win32con
    import win32net
    import wmi

    HAS_WIN32NET_MODS = True
except ImportError:
    HAS_WIN32NET_MODS = False

# Set up logging
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "system"


def __virtual__():
    """
    Only works on Windows Systems with Win32 Modules
    """
    if not salt.utils.platform.is_windows():
        return False, "Module win_system: Requires Windows"

    if not HAS_WIN32NET_MODS:
        return False, "Module win_system: Missing win32 modules"

    return __virtualname__


def _convert_minutes_seconds(timeout, in_seconds=False):
    """
    convert timeout to seconds
    """
    return timeout if in_seconds else timeout * 60


def _convert_date_time_string(dt_string):
    """
    convert string to date time object
    """
    dt_string = dt_string.split(".")[0]
    dt_obj = datetime.strptime(dt_string, "%Y%m%d%H%M%S")
    return dt_obj.strftime("%Y-%m-%d %H:%M:%S")


def halt(timeout=5, in_seconds=False):
    """
    Halt a running system.

    Args:

        timeout (:obj:`int`, optional):
            Number of minutes/seconds before halting the system.

            Default is 5 minutes.

        in_seconds (:obj:`bool`, optional):
            Whether to treat timeout as seconds or minutes.

            Default is ``True``.

            .. versionadded:: 2015.8.0

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.halt 5 True
    """
    return shutdown(timeout=timeout, in_seconds=in_seconds)


def init(runlevel):  # pylint: disable=unused-argument
    """
    Change the system runlevel on sysV compatible systems. Not applicable to
    Windows

    CLI Example:

    .. code-block:: bash

        salt '*' system.init 3
    """
    # cmd = ['init', runlevel]
    # ret = __salt__['cmd.run'](cmd, python_shell=False)
    # return ret

    # TODO: Create a mapping of runlevels to  # pylint: disable=fixme
    #       corresponding Windows actions

    return "Not implemented on Windows at this time."


def poweroff(timeout=5, in_seconds=False):
    """
    Power off a running system.

    Args:

        timeout (:obj:`int`, optional):
            Number of minutes/seconds before powering off the system.

            Default is 5 minutes.

        in_seconds (:obj:`bool`, optional):
            Whether to treat timeout as seconds or minutes.

            .. versionadded:: 2015.8.0

            Default is ``False``.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.poweroff 5
    """
    return shutdown(timeout=timeout, in_seconds=in_seconds)


def reboot(
    timeout=5,
    in_seconds=False,
    wait_for_reboot=False,  # pylint: disable=redefined-outer-name
    only_on_pending_reboot=False,
):
    """
    Reboot a running system.

    Args:

        timeout (:obj:`int`, optional):
            The number of minutes/seconds before rebooting the system. Use of
            minutes or seconds depends on the value of ``in_seconds``.

            Default is 5 minutes.

        in_seconds (:obj:`bool`, optional):
            ``True`` will cause the ``timeout`` parameter to be in seconds.
            ``False`` will be in minutes.

            Default is ``False``.

            .. versionadded:: 2015.8.0

        wait_for_reboot (:obj:`bool`, optional):
            ``True`` will sleep for timeout + 30 seconds after reboot has been
            initiated. This is useful for use in a highstate. For example, you
            may have states that you want to apply only after the reboot.

            Default is ``False``.

            .. versionadded:: 2015.8.0

        only_on_pending_reboot (:obj:`bool`, optional):
            If this is set to ``True``, then the reboot will only proceed
            if the system reports a pending reboot. Setting this parameter to
            ``True`` could be useful when calling this function from a final
            housekeeping state intended to be executed at the end of a state run
            (using *order: last*).

            Default is ``False``.

    Returns:
        bool: ``True`` if successful (a reboot will occur), otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.reboot 5
        salt '*' system.reboot 5 True

    Invoking this function from a final housekeeping state:

    .. code-block:: yaml

        final_housekeeping:
          module.run:
            - name: system.reboot
            - only_on_pending_reboot: True
            - order: last
    """
    ret = shutdown(
        timeout=timeout,
        reboot=True,
        in_seconds=in_seconds,
        only_on_pending_reboot=only_on_pending_reboot,
    )

    if wait_for_reboot:
        seconds = _convert_minutes_seconds(timeout, in_seconds)
        time.sleep(seconds + 30)

    return ret


def shutdown(
    message=None,
    timeout=5,
    force_close=True,
    reboot=False,  # pylint: disable=redefined-outer-name
    in_seconds=False,
    only_on_pending_reboot=False,
):
    """
    Shutdown a running system.

    Args:

        message (:obj:`str`, optional):
            The message to display to the user before shutting down.

            Default is ``None``.

        timeout (:obj:`int`, optional):
            The length of time (in seconds) that the shutdown dialog box should
            be displayed. While this dialog box is displayed, the shutdown can
            be aborted using the ``system.shutdown_abort`` function.

            If timeout is not zero, InitiateSystemShutdown displays a dialog box
            on the specified computer. The dialog box displays the name of the
            user who called the function, the message specified by the lpMessage
            parameter, and prompts the user to log off. The dialog box beeps
            when it is created and remains on top of other windows (system
            modal). The dialog box can be moved but not closed. A timer counts
            down the remaining time before the shutdown occurs.

            If timeout is zero, the computer shuts down immediately without
            displaying the dialog box and cannot be stopped by
            ``system.shutdown_abort``.

            Default is 5 minutes

        in_seconds (:obj:`bool`, optional):
            ``True`` will cause the ``timeout`` parameter to be in seconds.
             ``False`` will be in minutes.

             Default is ``False``.

            .. versionadded:: 2015.8.0

        force_close (:obj:`bool`, optional):
            ``True`` will force close all open applications. ``False`` will
            display a dialog box instructing the user to close open
            applications.

            Default is ``True``.

        reboot (:obj:`bool`, optional):
            ``True`` restarts the computer immediately after shutdown. ``False``
            powers down the system.

            Default is ``False``.

        only_on_pending_reboot (:obj:`bool`, optional):
            If ``True`` the shutdown will only proceed if there is a reboot
            pending. ``False`` will shutdown the system. To optionally shutdown
            in a highstate, consider using the shutdown state instead of this
            module.

            Default is ``False``.

    Returns:
        bool: ``True`` if successful (a shutdown or reboot will occur),
            otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' system.shutdown 'System will shutdown in 5 minutes'
    """
    timeout = _convert_minutes_seconds(timeout, in_seconds)

    if only_on_pending_reboot and not get_pending_reboot():
        return False

    if message and not isinstance(message, str):
        message = message.decode("utf-8")
    try:
        win32api.InitiateSystemShutdown(
            "127.0.0.1", message, timeout, force_close, reboot
        )
        return True
    except pywintypes.error as exc:
        (number, context, message) = exc.args
        log.error("Failed to shutdown the system")
        log.error("nbr: %s", number)
        log.error("ctx: %s", context)
        log.error("msg: %s", message)
        return False


def shutdown_hard():
    """
    Shutdown a running system with no timeout or warning.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.shutdown_hard
    """
    return shutdown(timeout=0)


def shutdown_abort():
    """
    Abort a shutdown. Only available while the dialog box is being
    displayed to the user. Once the shutdown has initiated, it cannot be
    aborted.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.shutdown_abort
    """
    try:
        win32api.AbortSystemShutdown("127.0.0.1")
        return True
    except pywintypes.error as exc:
        (number, context, message) = exc.args
        log.error("Failed to abort system shutdown")
        log.error("nbr: %s", number)
        log.error("ctx: %s", context)
        log.error("msg: %s", message)
        return False


def lock():
    """
    Lock the workstation.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.lock
    """
    return windll.user32.LockWorkStation()


def set_computer_name(name):
    """
    Set the Windows computer name

    Args:

        name (str):
            The new name to give the computer. Requires a reboot to take effect.

    Returns:
        dict: Returns a dictionary containing the old and new names.
        bool: ``False`` if failed to set computer name

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.set_computer_name 'DavesComputer'
    """
    if windll.kernel32.SetComputerNameExW(
        win32con.ComputerNamePhysicalDnsHostname, name
    ):
        ret = {"Computer Name": {"Current": get_computer_name()}}
        pending = get_pending_computer_name()
        if pending not in (None, False):
            ret["Computer Name"]["Pending"] = pending
        return ret

    return False


def get_pending_computer_name():
    """
    Get a pending computer name. If the computer name has been changed, and the
    change is pending a system reboot, this function will return the pending
    computer name. Otherwise, ``None`` will be returned. If there was an error
    retrieving the pending computer name, ``False`` will be returned, and an
    error message will be logged to the minion log.

    Returns:
        str: Returns the pending name if pending restart. Returns ``None`` if
            not pending restart.

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.get_pending_computer_name
    """
    return salt.utils.win_system.get_pending_computer_name()


def get_computer_name():
    """
    Get the Windows computer name

    Returns:
        str: Returns the computer name if found. Otherwise, returns ``False``.

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.get_computer_name
    """
    return salt.utils.win_system.get_computer_name()


def set_computer_desc(desc=None):
    """
    Set the Windows computer description.

    Args:

        desc (:obj:`str`, optional): The computer description.

            Default is ``None``.

    Returns:
        str: Description if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.set_computer_desc 'This computer belongs to Dave!'
    """
    # Make sure the system exists
    # Return an object containing current information array for the computer
    system_info = win32net.NetServerGetInfo(None, 101)

    # If desc is passed, decode it for unicode
    if desc is None:
        return False

    system_info["comment"] = desc

    # Apply new settings
    try:
        win32net.NetServerSetInfo(None, 101, system_info)
    except win32net.error as exc:
        (number, context, message) = exc.args
        log.error("Failed to update system")
        log.error("nbr: %s", number)
        log.error("ctx: %s", context)
        log.error("msg: %s", message)
        return False

    return {"Computer Description": get_computer_desc()}


# pylint: disable=invalid-name
set_computer_description = salt.utils.functools.alias_function(
    set_computer_desc, "set_computer_description"
)
# pylint: enable=invalid-name


def get_system_info():
    """
    Get system information.

    .. note::

        Not all system info is available across all versions of Windows. If it
        is not available on an older version, it will be skipped

    Returns:
        dict: Dictionary containing information about the system to include
        name, description, version, etc...

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.get_system_info
    """

    def byte_calc(val):
        val = float(val)
        if val < 2**10:
            return f"{val:.3f}B"
        elif val < 2**20:
            return f"{val / 2**10:.3f}KB"
        elif val < 2**30:
            return f"{val / 2**20:.3f}MB"
        elif val < 2**40:
            return f"{val / 2**30:.3f}GB"
        else:
            return f"{val / 2**40:.3f}TB"

    # Lookup dicts for Win32_OperatingSystem
    os_type = {1: "Work Station", 2: "Domain Controller", 3: "Server"}
    # lookup dicts for Win32_ComputerSystem
    domain_role = {
        0: "Standalone Workstation",
        1: "Member Workstation",
        2: "Standalone Server",
        3: "Member Server",
        4: "Backup Domain Controller",
        5: "Primary Domain Controller",
    }
    warning_states = {
        1: "Other",
        2: "Unknown",
        3: "Safe",
        4: "Warning",
        5: "Critical",
        6: "Non-recoverable",
    }
    pc_system_types = {
        0: "Unspecified",
        1: "Desktop",
        2: "Mobile",
        3: "Workstation",
        4: "Enterprise Server",
        5: "SOHO Server",
        6: "Appliance PC",
        7: "Performance Server",
        8: "Maximum",
    }

    # Connect to WMI
    with salt.utils.winapi.Com():
        conn = wmi.WMI()

        system = conn.Win32_OperatingSystem()[0]
        ret = {
            "name": get_computer_name(),
            "description": system.Description,
            "install_date": system.InstallDate,
            "last_boot": system.LastBootUpTime,
            "os_manufacturer": system.Manufacturer,
            "os_name": system.Caption,
            "users": system.NumberOfUsers,
            "organization": system.Organization,
            "os_architecture": system.OSArchitecture,
            "primary": system.Primary,
            "os_type": os_type[system.ProductType],
            "registered_user": system.RegisteredUser,
            "system_directory": system.SystemDirectory,
            "system_drive": system.SystemDrive,
            "os_version": system.Version,
            "windows_directory": system.WindowsDirectory,
        }
        # Must get chassis_sku_number this way for backwards compatibility
        # system.ChassisSKUNumber is only available on Windows 10/2016 and newer
        product = conn.Win32_ComputerSystemProduct()[0]
        ret.update({"chassis_sku_number": product.SKUNumber})
        system = conn.Win32_ComputerSystem()[0]
        # Get pc_system_type depending on Windows version
        if platform.release() in ["Vista", "7", "8"]:
            # Types for Vista, 7, and 8
            pc_system_type = pc_system_types[system.PCSystemType]
        else:
            # New types were added with 8.1 and newer
            pc_system_types.update({8: "Slate", 9: "Maximum"})
            pc_system_type = pc_system_types[system.PCSystemType]
        ret.update(
            {
                "bootup_state": system.BootupState,
                "caption": system.Caption,
                "chassis_bootup_state": warning_states[system.ChassisBootupState],
                "dns_hostname": system.DNSHostname,
                "domain": system.Domain,
                "domain_role": domain_role[system.DomainRole],
                "hardware_manufacturer": system.Manufacturer,
                "hardware_model": system.Model,
                "network_server_mode_enabled": system.NetworkServerModeEnabled,
                "part_of_domain": system.PartOfDomain,
                "pc_system_type": pc_system_type,
                "power_state": system.PowerState,
                "status": system.Status,
                "system_type": system.SystemType,
                "total_physical_memory": byte_calc(system.TotalPhysicalMemory),
                "total_physical_memory_raw": system.TotalPhysicalMemory,
                "thermal_state": warning_states[system.ThermalState],
                "workgroup": system.Workgroup,
            }
        )
        # Get processor information
        processors = conn.Win32_Processor()
        ret["processors"] = 0
        ret["processors_logical"] = 0
        ret["processor_cores"] = 0
        ret["processor_cores_enabled"] = 0
        ret["processor_manufacturer"] = processors[0].Manufacturer
        ret["processor_max_clock_speed"] = str(processors[0].MaxClockSpeed) + "MHz"
        for processor in processors:
            ret["processors"] += 1
            ret["processors_logical"] += processor.NumberOfLogicalProcessors
            ret["processor_cores"] += processor.NumberOfCores
            # Older versions of Windows do not have the NumberOfEnabledCore
            # property. In that case, we'll just skip it
            try:
                ret["processor_cores_enabled"] += processor.NumberOfEnabledCore
            except (AttributeError, TypeError):
                pass
        if ret["processor_cores_enabled"] == 0:
            ret.pop("processor_cores_enabled", False)

        bios = conn.Win32_BIOS()[0]
        ret.update(
            {
                "hardware_serial": bios.SerialNumber,
                "bios_manufacturer": bios.Manufacturer,
                "bios_version": bios.Version,
                "bios_details": bios.BIOSVersion,
                "bios_caption": bios.Caption,
                "bios_description": bios.Description,
            }
        )
        ret["install_date"] = _convert_date_time_string(ret["install_date"])
        ret["last_boot"] = _convert_date_time_string(ret["last_boot"])
    return ret


def get_computer_desc():
    """
    Get the Windows computer description.

    Returns:
        str: Returns the computer description if found. Otherwise, returns
            ``False``.

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.get_computer_desc
    """
    desc = get_system_info()["description"]
    return False if desc is None else desc


# pylint: disable=invalid-name
get_computer_description = salt.utils.functools.alias_function(
    get_computer_desc, "get_computer_description"
)
# pylint: enable=invalid-name


def get_hostname():
    """
    Get the hostname of the windows minion.

    .. versionadded:: 2016.3.0

    Returns:
        str: Returns the hostname of the windows minion

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.get_hostname
    """
    cmd = "hostname"
    ret = __salt__["cmd.run"](cmd=cmd)
    return ret


def set_hostname(hostname):
    """
    Set the hostname of the windows minion, requires a restart before this will
    be updated.

    .. versionadded:: 2016.3.0

    Args:

        hostname (str): The hostname to set.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.set_hostname newhostname
    """
    with salt.utils.winapi.Com():
        conn = wmi.WMI()
        comp = conn.Win32_ComputerSystem()[0]
        return comp.Rename(Name=hostname)


def join_domain(
    domain,
    username=None,
    password=None,
    account_ou=None,
    account_exists=False,
    restart=False,
):
    """
    Join a computer to an Active Directory domain. Requires a reboot.

    Args:

        domain (str):
            The domain to which the computer should be joined, e.g.
            ``example.com``.

        username (:obj:`str`, optional):
            Username of an account which is authorized to join computers to the
            specified domain. Needs to be either fully qualified like
            ``user@domain.tld`` or simply ``user``.

            Default is ``None``.

        password (:obj:`str`, optional):
            Password of the specified user.

            Default is ``None``.

        account_ou (:obj:`str`, optional):
            The DN of the OU below which the account for this computer should be
            created when joining the domain, e.g.
            ``ou=computers,ou=departm_432,dc=my-company,dc=com``.

            Default is ``None``.

        account_exists (:obj:`bool`, optional):
            If set to ``True`` the computer will only join the domain if the
            account already exists. If set to ``False``, the computer account
            will be created if it does not exist, otherwise it will use the
            existing account.

            Default is ``False``.

        restart (:obj:`bool`, optional):
            ``True`` will restart the computer after a successful join.

             Default is ``False``.

            .. versionadded:: 2015.5.7,2015.8.2

    Returns:
        dict: Returns a dictionary if successful, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.join_domain domain='domain.tld' \\
                         username='joinuser' password='joinpassword' \\
                         account_ou='ou=clients,ou=org,dc=domain,dc=tld' \\
                         account_exists=False, restart=True
    """
    status = get_domain_workgroup()
    if "Domain" in status:
        if status["Domain"] == domain:
            return f"Already joined to {domain}"

    if username and "\\" not in username and "@" not in username:
        username = f"{username}@{domain}"

    if username and password is None:
        return "Must specify a password if you pass a username"

    # remove any escape characters
    if isinstance(account_ou, str):
        account_ou = account_ou.split("\\")
        account_ou = "".join(account_ou)

    err = _join_domain(
        domain=domain,
        username=username,
        password=password,
        account_ou=account_ou,
        account_exists=account_exists,
    )

    if not err:
        ret = {"Domain": domain, "Restart": False}
        if restart:
            ret["Restart"] = reboot()
        return ret

    raise CommandExecutionError(win32api.FormatMessage(err).rstrip())


def _join_domain(
    domain, username=None, password=None, account_ou=None, account_exists=False
):
    """
    Helper function to join the domain.

    Args:

        domain (str):
            The domain to which the computer should be joined, e.g.
            ``example.com``.

        username (:obj:`str`, optional):
            Username of an account which is authorized to join computers to the
            specified domain. Need to be either fully qualified like
            ``user@domain.tld`` or simply ``user``.

            Default is ``None``.

        password (:obj:`str`, optional): Password of the specified user.

            Default is ``None``.

        account_ou (:obj:`str`, optional):
            The DN of the OU below which the account for this computer should be
            created when joining the domain, e.g.
            ``ou=computers,ou=departm_432,dc=my-company,dc=com``.

            Default is ``None``.

        account_exists (:obj:`bool`, optional):
            If set to ``True`` the computer will only join the domain if the
            account already exists. If set to ``False``, the computer account
            will be created if it does not exist, otherwise it will use the
            existing account.

            Default is ``False``.

    Returns:
        int: Representing the success of failure of the domain join
    """
    NETSETUP_JOIN_DOMAIN = 0x1  # pylint: disable=invalid-name
    NETSETUP_ACCOUNT_CREATE = 0x2  # pylint: disable=invalid-name
    NETSETUP_DOMAIN_JOIN_IF_JOINED = 0x20  # pylint: disable=invalid-name
    NETSETUP_JOIN_WITH_NEW_NAME = 0x400  # pylint: disable=invalid-name

    join_options = 0x0
    join_options |= NETSETUP_JOIN_DOMAIN
    join_options |= NETSETUP_DOMAIN_JOIN_IF_JOINED
    join_options |= NETSETUP_JOIN_WITH_NEW_NAME
    if not account_exists:
        join_options |= NETSETUP_ACCOUNT_CREATE

    with salt.utils.winapi.Com():
        conn = wmi.WMI()
        comp = conn.Win32_ComputerSystem()[0]

        # Return the results of the command as an error
        # JoinDomainOrWorkgroup returns a strangely formatted value that looks like
        # (0,) so return the first item
        return comp.JoinDomainOrWorkgroup(
            Name=domain,
            Password=password,
            UserName=username,
            AccountOU=account_ou,
            FJoinOptions=join_options,
        )[0]


def unjoin_domain(
    username=None,
    password=None,
    domain=None,
    workgroup="WORKGROUP",
    disable=False,
    restart=False,
):
    # pylint: disable=anomalous-backslash-in-string
    """
    Unjoin a computer from an Active Directory Domain. Requires a restart.

    Args:

        username (:obj:`str`, optional):
            Username of an account which is authorized to manage computer
            accounts on the domain. Needs to be a fully qualified name like
            ``user@domain.tld`` or ``domain.tld\\user``. If the domain is not
            specified, the passed domain will be used. If the computer account
            doesn't need to be disabled after the computer is unjoined, this can
            be ``None``.

            Default is ``None``.

        password (:obj:`str`, optional):
            The password of the specified user.

            Default is ``None``.

        domain (:obj:`str`, optional):
            The domain from which to unjoin the computer. Can be ``None``.

            Default is ``None``.

        workgroup (:obj:`str`, optional):
            The workgroup to join the computer to.

            Default is ``WORKGROUP``.

            .. versionadded:: 2015.5.7,2015.8.2

        disable (:obj:`bool`, optional):
            ``True`` to disable the computer account in Active Directory.

            Default is ``False``

        restart (:obj:`bool`, optional):
            ``True`` will restart the computer after successful unjoin.

            Default is ``False``.

            .. versionadded:: 2015.5.7,2015.8.2

    Returns:
        dict: Returns a dictionary if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.unjoin_domain restart=True

        salt 'minion-id' system.unjoin_domain username='unjoinuser' \\
                         password='unjoinpassword' disable=True \\
                         restart=True
    """
    # pylint: enable=anomalous-backslash-in-string
    status = get_domain_workgroup()
    if "Workgroup" in status:
        if status["Workgroup"] == workgroup:
            return f"Already joined to {workgroup}"

    if username and "\\" not in username and "@" not in username:
        if domain:
            username = f"{username}@{domain}"
        else:
            return "Must specify domain if not supplied in username"

    if username and password is None:
        return "Must specify a password if you pass a username"

    NETSETUP_ACCT_DELETE = 0x4  # pylint: disable=invalid-name

    unjoin_options = 0x0
    if disable:
        unjoin_options |= NETSETUP_ACCT_DELETE

    with salt.utils.winapi.Com():
        conn = wmi.WMI()
        comp = conn.Win32_ComputerSystem()[0]
        err = comp.UnjoinDomainOrWorkgroup(
            Password=password, UserName=username, FUnjoinOptions=unjoin_options
        )

        # you have to do this because UnjoinDomainOrWorkgroup returns a
        # strangely formatted value that looks like (0,)
        if not err[0]:
            err = comp.JoinDomainOrWorkgroup(Name=workgroup)
            if not err[0]:
                ret = {"Workgroup": workgroup, "Restart": False}
                if restart:
                    ret["Restart"] = reboot()

                return ret
            else:
                log.error(win32api.FormatMessage(err[0]).rstrip())
                log.error("Failed to unjoin the computer from %s", workgroup)
                return False
        else:
            log.error(win32api.FormatMessage(err[0]).rstrip())
            log.error("Failed to unjoin computer from %s", status["Domain"])
            return False


def get_domain_workgroup():
    """
    Get the domain or workgroup the computer belongs to.

    .. versionadded:: 2015.5.7
    .. versionadded:: 2015.8.2

    Returns:
        str: The name of the domain or workgroup.

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.get_domain_workgroup
    """
    with salt.utils.winapi.Com():
        conn = wmi.WMI()
        for computer in conn.Win32_ComputerSystem():
            if computer.PartOfDomain:
                return {"Domain": computer.Domain}
            else:
                return {"Workgroup": computer.Domain}


def set_domain_workgroup(workgroup):
    """
    Set the domain or workgroup the computer belongs to.

    .. versionadded:: 3001

    Args:

        workgroup (str):
            The name of the domain or workgroup.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.set_domain_workgroup LOCAL
    """
    # Initialize COM
    with salt.utils.winapi.Com():
        # Grab the first Win32_ComputerSystem object from wmi
        conn = wmi.WMI()
        comp = conn.Win32_ComputerSystem()[0]

        # Now we can join the new workgroup
        res = comp.JoinDomainOrWorkgroup(Name=workgroup.upper())

    return True if not res[0] else False


def _try_parse_datetime(time_str, fmts):
    """
    A helper function that attempts to parse the input time_str as a date.

    Args:

        time_str (str): A string representing the time

        fmts (list): A list of date format strings

    Returns:
        datetime: Returns a datetime object if parsed properly, otherwise None
    """
    result = None
    for fmt in fmts:
        try:
            result = datetime.strptime(time_str, fmt)
            break
        except ValueError:
            pass
    return result


def get_system_time():
    """
    Get the system time.

    Returns:
        str: Returns the system time in HH:MM:SS AM/PM format.

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.get_system_time
    """
    now = win32api.GetLocalTime()
    meridian = "AM"
    hours = int(now[4])
    if hours == 12:
        meridian = "PM"
    elif hours == 0:
        hours = 12
    elif hours > 12:
        hours = hours - 12
        meridian = "PM"
    return f"{hours:02d}:{now[5]:02d}:{now[6]:02d} {meridian}"


def set_system_time(newtime):
    """
    Set the system time.

    Args:

        newtime (str):
            The time to set. Can be any of the following formats:

            - HH:MM:SS AM/PM
            - HH:MM AM/PM
            - HH:MM:SS (24 hour)
            - HH:MM (24 hour)

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt 'minion-id' system.set_system_time 12:01
    """
    # Get date/time object from newtime
    fmts = ["%I:%M:%S %p", "%I:%M %p", "%H:%M:%S", "%H:%M"]
    dt_obj = _try_parse_datetime(newtime, fmts)
    if dt_obj is None:
        return False

    # Set time using set_system_date_time()
    return set_system_date_time(
        hours=dt_obj.hour, minutes=dt_obj.minute, seconds=dt_obj.second
    )


def set_system_date_time(
    years=None, months=None, days=None, hours=None, minutes=None, seconds=None
):
    """
    Set the system date and time. Each argument is an element of the date, but
    not required. If an element is not passed, the current system value for that
    element will be used. For example, if you don't pass the year, the current
    system year will be used. (Used by set_system_date and set_system_time)

    Args:

        years (:obj:`int`, optional): Years digit, ie: 2015

            Default is ``None``.

        months (:obj:`int`, optional): Months digit: 1 - 12

            Default is ``None``.

        days (:obj:`int`, optional): Days digit: 1 - 31

            Default is ``None``.

        hours (:obj:`int`, optional): Hours digit: 0 - 23

            Default is ``None``.

        minutes (:obj:`int`, optional): Minutes digit: 0 - 59

            Default is ``None``.

        seconds (:obj:`int`, optional): Seconds digit: 0 - 59

            Default is ``None``.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.set_system_date_ time 2015 5 12 11 37 53
    """
    # Get the current date/time
    try:
        date_time = win32api.GetLocalTime()
    except win32api.error as exc:
        (number, context, message) = exc.args
        log.error("Failed to get local time")
        log.error("nbr: %s", number)
        log.error("ctx: %s", context)
        log.error("msg: %s", message)
        return False

    # Check for passed values. If not passed, use current values
    if years is None:
        years = date_time[0]
    if months is None:
        months = date_time[1]
    if days is None:
        days = date_time[3]
    if hours is None:
        hours = date_time[4]
    if minutes is None:
        minutes = date_time[5]
    if seconds is None:
        seconds = date_time[6]

    try:

        class SYSTEMTIME(ctypes.Structure):
            _fields_ = [
                ("wYear", ctypes.c_int16),
                ("wMonth", ctypes.c_int16),
                ("wDayOfWeek", ctypes.c_int16),
                ("wDay", ctypes.c_int16),
                ("wHour", ctypes.c_int16),
                ("wMinute", ctypes.c_int16),
                ("wSecond", ctypes.c_int16),
                ("wMilliseconds", ctypes.c_int16),
            ]

        system_time = SYSTEMTIME()
        # pylint: disable=invalid-name
        system_time.wYear = int(years)
        system_time.wMonth = int(months)
        system_time.wDay = int(days)
        system_time.wHour = int(hours)
        system_time.wMinute = int(minutes)
        system_time.wSecond = int(seconds)
        # pylint: enable=invalid-name
        system_time_ptr = ctypes.pointer(system_time)
        succeeded = ctypes.windll.kernel32.SetLocalTime(system_time_ptr)
        if succeeded != 0:
            return True
        else:
            log.error("Failed to set local time")
            raise CommandExecutionError(win32api.FormatMessage(succeeded).rstrip())
    except OSError as err:
        log.error("Failed to set local time")
        raise CommandExecutionError(err)


def get_system_date():
    """
    Get the Windows system date

    Returns:
        str: Returns the system date.

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_system_date
    """
    now = win32api.GetLocalTime()
    return f"{now[1]:02d}/{now[3]:02d}/{now[0]:04d}"


def set_system_date(newdate):
    """
    Set the Windows system date. Use <mm-dd-yy> format for the date.

    Args:

        newdate (str):
            The date to set. Can be any of the following formats

            - YYYY-MM-DD
            - MM-DD-YYYY
            - MM-DD-YY
            - MM/DD/YYYY
            - MM/DD/YY
            - YYYY/MM/DD

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.set_system_date '03-28-13'
    """
    fmts = ["%Y-%m-%d", "%m-%d-%Y", "%m-%d-%y", "%m/%d/%Y", "%m/%d/%y", "%Y/%m/%d"]
    # Get date/time object from newdate
    dt_obj = _try_parse_datetime(newdate, fmts)
    if dt_obj is None:
        return False

    # Set time using set_system_date_time()
    return set_system_date_time(years=dt_obj.year, months=dt_obj.month, days=dt_obj.day)


def start_time_service():
    """
    Start the Windows time service.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.start_time_service
    """
    return __salt__["service.start"]("w32time")


def stop_time_service():
    """
    Stop the Windows time service.

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.stop_time_service
    """
    return __salt__["service.stop"]("w32time")


def get_pending_component_servicing():
    """
    Determine whether there are pending Component Based Servicing tasks that
    require a reboot.

    .. versionadded:: 2016.11.0

    Returns:
        bool: ``True`` if there are pending Component Based Servicing tasks,
        otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_pending_component_servicing
    """
    return salt.utils.win_system.get_pending_component_servicing()


def get_pending_domain_join():
    """
    Determine whether there is a pending domain join action that requires a
    reboot.

    .. versionadded:: 2016.11.0

    Returns:
        bool: ``True`` if there is a pending domain join action, otherwise
        ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_pending_domain_join
    """
    return salt.utils.win_system.get_pending_domain_join()


def get_pending_file_rename():
    """
    Determine whether there are pending file rename operations that require a
    reboot.

    .. versionadded:: 2016.11.0

    Returns:
        bool: ``True`` if there are pending file rename operations, otherwise
        ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_pending_file_rename
    """
    return salt.utils.win_system.get_pending_file_rename()


def get_pending_servermanager():
    """
    Determine whether there are pending Server Manager tasks that require a
    reboot.

    .. versionadded:: 2016.11.0

    Returns:
        bool: ``True`` if there are pending Server Manager tasks, otherwise
        ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_pending_servermanager
    """
    return salt.utils.win_system.get_pending_servermanager()


def get_pending_update():
    """
    Determine whether there are pending updates that require a reboot.

    .. versionadded:: 2016.11.0

    Returns:
        bool: ``True`` if there are pending updates, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_pending_update
    """
    return salt.utils.win_system.get_pending_update()


def set_reboot_required_witnessed():
    r"""
    This function is used to remember that an event indicating that a reboot is
    required was witnessed. This function relies on the salt-minion's ability to
    create the following volatile registry key in the *HKLM* hive:

       *SYSTEM\\CurrentControlSet\\Services\\salt-minion\\Volatile-Data*

    Because this registry key is volatile, it will not persist beyond the
    current boot session. Also, in the scope of this key, the name *'Reboot
    required'* will be assigned the value of *1*.

    For the time being, this function is being used whenever an install
    completes with exit code 3010 and can be extended where appropriate in the
    future.

    .. versionadded:: 2016.11.0

    Returns:
        bool: ``True`` if successful, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.set_reboot_required_witnessed
    """
    return salt.utils.win_system.set_reboot_required_witnessed()


def get_reboot_required_witnessed():
    """
    Determine if at any time during the current boot session the salt minion
    witnessed an event indicating that a reboot is required.

    This function will return ``True`` if an install completed with exit
    code 3010 during the current boot session and can be extended where
    appropriate in the future.

    .. versionadded:: 2016.11.0

    Returns:
        bool: ``True`` if the ``Requires reboot`` registry flag is set to ``1``,
        otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_reboot_required_witnessed

    """
    return salt.utils.win_system.get_reboot_required_witnessed()


def get_pending_reboot():
    """
    Determine whether there is a reboot pending.

    .. versionadded:: 2016.11.0

    Returns:
        bool: ``True`` if the system is pending reboot, otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_pending_reboot
    """
    return salt.utils.win_system.get_pending_reboot()


def get_pending_reboot_details():
    """
    Determine which check is signalling that the system is pending a reboot.
    Useful in determining why your system is signalling that it needs a reboot.

    .. versionadded:: 3001

    Returns:
        dict: A dictionary of the results of each system that would indicate a
            pending reboot.

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_pending_reboot_details
    """
    return salt.utils.win_system.get_pending_reboot_details()


def get_pending_windows_update():
    """
    Check the Windows Update system for a pending reboot state.

    This leverages the Windows Update System to determine if the system is
    pending a reboot.

    .. versionadded:: 3001

    Returns:
        bool: ``True`` if the Windows Update system reports a pending update,
            otherwise ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' system.get_pending_windows_update
    """
    return salt.utils.win_system.get_pending_windows_update()



"""
Module for managing Windows Updates using the Windows Update Agent.

List updates on the system using the following functions:

- :py:func:`win_wua.available <salt.modules.win_wua.available>`
- :py:func:`win_wua.list <salt.modules.win_wua.list_>`

This is an easy way to find additional information about updates available to
to the system, such as the GUID, KB number, or description.

Once you have the GUID or a KB number for the update you can get information
about the update, download, install, or uninstall it using these functions:

- :py:func:`win_wua.get <salt.modules.win_wua.get>`
- :py:func:`win_wua.download <salt.modules.win_wua.download>`
- :py:func:`win_wua.install <salt.modules.win_wua.install>`
- :py:func:`win_wua.uninstall <salt.modules.win_wua.uninstall>`

The get function expects a name in the form of a GUID, KB, or Title and should
return information about a single update. The other functions accept either a
single item or a list of items for downloading/installing/uninstalling a
specific list of items.

The :py:func:`win_wua.list <salt.modules.win_wua.list_>` and
:py:func:`win_wua.get <salt.modules.win_wua.get>` functions are utility
functions. In addition to returning information about updates they can also
download and install updates by setting ``download=True`` or ``install=True``.
So, with py:func:`win_wua.list <salt.modules.win_wua.list_>` for example, you
could run the function with the filters you want to see what is available. Then
just add ``install=True`` to install everything on that list.

If you want to download, install, or uninstall specific updates, use
:py:func:`win_wua.download <salt.modules.win_wua.download>`,
:py:func:`win_wua.install <salt.modules.win_wua.install>`, or
:py:func:`win_wua.uninstall <salt.modules.win_wua.uninstall>`. To update your
system with the latest updates use :py:func:`win_wua.list
<salt.modules.win_wua.list_>` and set ``install=True``

You can also adjust the Windows Update settings using the
:py:func:`win_wua.set_wu_settings <salt.modules.win_wua.set_wu_settings>`
function. This function is only supported on the following operating systems:

- Windows Vista / Server 2008
- Windows 7 / Server 2008R2
- Windows 8 / Server 2012
- Windows 8.1 / Server 2012R2

As of Windows 10 and Windows Server 2016, the ability to modify the Windows
Update settings has been restricted. The settings can be modified in the Local
Group Policy using the ``lgpo`` module.

.. versionadded:: 2015.8.0

:depends: salt.utils.win_update
"""

import logging

import salt.utils.platform
import salt.utils.win_service
import salt.utils.win_update
import salt.utils.winapi
from salt.exceptions import CommandExecutionError

try:
    import win32com.client

    HAS_PYWIN32 = True
except ImportError:
    HAS_PYWIN32 = False

log = logging.getLogger(__name__)

__func_alias__ = {
    "list_": "list",
}


def __virtual__():
    """
    Only works on Windows systems with PyWin32
    """
    if not salt.utils.platform.is_windows():
        return False, "WUA: Only available on Windows systems"

    if not HAS_PYWIN32:
        return False, "WUA: Requires PyWin32 libraries"

    if not salt.utils.win_update.HAS_PYWIN32:
        return False, "WUA: Missing Libraries required by salt.utils.win_update"

    if salt.utils.win_service.info("wuauserv")["StartType"] == "Disabled":
        return (
            False,
            "WUA: The Windows Update service (wuauserv) must not be disabled",
        )

    if salt.utils.win_service.info("msiserver")["StartType"] == "Disabled":
        return (
            False,
            "WUA: The Windows Installer service (msiserver) must not be disabled",
        )

    if salt.utils.win_service.info("BITS")["StartType"] == "Disabled":
        return (
            False,
            "WUA: The Background Intelligent Transfer service (bits) must not "
            "be disabled",
        )

    if salt.utils.win_service.info("CryptSvc")["StartType"] == "Disabled":
        return (
            False,
            "WUA: The Cryptographic Services service (CryptSvc) must not be disabled",
        )

    if salt.utils.win_service.info("TrustedInstaller")["StartType"] == "Disabled":
        return (
            False,
            "WUA: The Windows Module Installer service (TrustedInstaller) must "
            "not be disabled",
        )

    return True


def available(
    software=True,
    drivers=True,
    summary=False,
    skip_installed=True,
    skip_hidden=True,
    skip_mandatory=False,
    skip_reboot=False,
    categories=None,
    severities=None,
    online=True,
):
    """
    .. versionadded:: 2017.7.0

    List updates that match the passed criteria. This allows for more filter
    options than :func:`list`. Good for finding a specific GUID or KB.

    Args:

        software (:obj:`bool`, optional):
            Include software updates in the results.

            Default is ``True``.

        drivers (:obj:`bool`, optional):
            Include driver updates in the results.

            Default is ``True``.

        summary (:obj:`bool`, optional):
            - ``True``: Return a summary of updates available for each category.
            - ``False`` (default): Return a detailed list of available updates.

            Default is ``False``.

        skip_installed (:obj:`bool`, optional):
            Skip updates that are already installed.

            Default is ``True``.

        skip_hidden (:obj:`bool`, optional):
            Skip updates that have been hidden.

            Default is ``True``.

        skip_mandatory (:obj:`bool`, optional):
            Skip mandatory updates.

            Default is ``False``.

        skip_reboot (:obj:`bool`, optional):
            Skip updates that require a reboot.

            Default is ``False``.

        categories (:obj:`list`, optional):
            Specify the categories to list. Must be passed as a list. All
            categories returned by default.

            Categories include the following:

            * Critical Updates
            * Definition Updates
            * Drivers (make sure you set ``drivers=True``)
            * Feature Packs
            * Security Updates
            * Update Rollups
            * Updates
            * Update Rollups
            * Windows 7
            * Windows 8.1
            * Windows 8.1 drivers
            * Windows 8.1 and later drivers
            * Windows Defender

            Default is ``None``.

        severities (:obj:`list`, optional):
            Specify the severities to include. Must be passed as a list. All
            severities returned by default.

            Severities include the following:

            * Critical
            * Important

            Default is ``None``.

        online (:obj:`bool`, optional):
            Tells the Windows Update Agent go online to update its local update
            database. ``True`` will go online. ``False`` will use the local
            update database as is. Default is ``True``

            .. versionadded:: 3001

            Default is ``True``.

    Returns:

        dict: Returns a dict containing either a summary or a list of updates:

        .. code-block:: cfg

            Dict of Updates:
            {'<GUID>': {
                'Title': <title>,
                'KB': <KB>,
                'GUID': <the globally unique identifier for the update>,
                'Description': <description>,
                'Downloaded': <has the update been downloaded>,
                'Installed': <has the update been installed>,
                'Mandatory': <is the update mandatory>,
                'UserInput': <is user input required>,
                'EULAAccepted': <has the EULA been accepted>,
                'Severity': <update severity>,
                'NeedsReboot': <is the update installed and awaiting reboot>,
                'RebootBehavior': <will the update require a reboot>,
                'Categories': [
                    '<category 1>',
                    '<category 2>',
                    ... ]
            }}

            Summary of Updates:
            {'Total': <total number of updates returned>,
             'Available': <updates that are not downloaded or installed>,
             'Downloaded': <updates that are downloaded but not installed>,
             'Installed': <updates installed (usually 0 unless installed=True)>,
             'Categories': {
                <category 1>: <total for that category>,
                <category 2>: <total for category 2>,
                ... }
            }

    CLI Examples:

    .. code-block:: bash

        # Normal Usage (list all software updates)
        salt '*' win_wua.available

        # List all updates with categories of Critical Updates and Drivers
        salt '*' win_wua.available categories='["Critical Updates","Drivers"]'

        # List all Critical Security Updates
        salt '*' win_wua.available categories='["Security Updates"]' severities='["Critical"]'

        # List all updates with a severity of Critical
        salt '*' win_wua.available severities='["Critical"]'

        # A summary of all available updates
        salt '*' win_wua.available summary=True

        # A summary of all Feature Packs and Windows 8.1 Updates
        salt '*' win_wua.available categories='["Feature Packs","Windows 8.1"]' summary=True
    """

    # Create a Windows Update Agent instance
    wua = salt.utils.win_update.WindowsUpdateAgent(online=online)

    # Look for available
    updates = wua.available(
        skip_hidden=skip_hidden,
        skip_installed=skip_installed,
        skip_mandatory=skip_mandatory,
        skip_reboot=skip_reboot,
        software=software,
        drivers=drivers,
        categories=categories,
        severities=severities,
    )

    # Return results as Summary or Details
    return updates.summary() if summary else updates.list()


def get(name, download=False, install=False, online=True):
    """
    .. versionadded:: 2017.7.0

    Returns details for the named update

    Args:

        name (str):
            The name of the update you're searching for. This can be the GUID, a
            KB number, or any part of the name of the update. GUIDs and KBs are
            preferred. Run ``list`` to get the GUID for the update you're
            looking for.

        download (:obj:`bool`, optional):
            Download the update returned by this function. Run this function
            first to see if the update exists, then set ``download=True`` to
            download the update.

            Default is ``False``.

        install (:obj:`bool`, optional):
            Install the update returned by this function. Run this function
            first to see if the update exists, then set ``install=True`` to
            install the update.

            Default is ``False``.

        online (:obj:`bool`, optional):
            Tells the Windows Update Agent go online to update its local update
            database. ``True`` will go online. ``False`` will use the local
            update database as is.

            Default is ``True``.

            .. versionadded:: 3001

    Returns:

        dict:
            Returns a dict containing a list of updates that match the name if
            download and install are both set to False. Should usually be a
            single update, but can return multiple if a partial name is given.

        If download or install is set to true it will return the results of the
        operation.

        .. code-block:: cfg

            Dict of Updates:
            {'<GUID>': {
                'Title': <title>,
                'KB': <KB>,
                'GUID': <the globally unique identifier for the update>,
                'Description': <description>,
                'Downloaded': <has the update been downloaded>,
                'Installed': <has the update been installed>,
                'Mandatory': <is the update mandatory>,
                'UserInput': <is user input required>,
                'EULAAccepted': <has the EULA been accepted>,
                'Severity': <update severity>,
                'NeedsReboot': <is the update installed and awaiting reboot>,
                'RebootBehavior': <will the update require a reboot>,
                'Categories': [
                    '<category 1>',
                    '<category 2>',
                    ... ]
            }}

    CLI Examples:

    .. code-block:: bash

        # Recommended Usage using GUID without braces
        # Use this to find the status of a specific update
        salt '*' win_wua.get 12345678-abcd-1234-abcd-1234567890ab

        # Use the following if you don't know the GUID:

        # Using a KB number
        # Not all updates have an associated KB
        salt '*' win_wua.get KB3030298

        # Using part or all of the name of the update
        # Could possibly return multiple results
        # Not all updates have an associated KB
        salt '*' win_wua.get 'Microsoft Camera Codec Pack'
    """
    # Create a Windows Update Agent instance
    wua = salt.utils.win_update.WindowsUpdateAgent(online=online)

    # Search for Update
    updates = wua.search(name)

    ret = {}

    # Download
    if download or install:
        ret["Download"] = wua.download(updates)

    # Install
    if install:
        ret["Install"] = wua.install(updates)

    return ret if ret else updates.list()


def list(
    software=True,
    drivers=False,
    summary=False,
    skip_installed=True,
    categories=None,
    severities=None,
    download=False,
    install=False,
    online=True,
):
    """
    .. versionadded:: 2017.7.0

    Returns a detailed list of available updates or a summary. If ``download``
    or ``install`` is ``True`` the same list will be downloaded and/or
    installed.

    Args:

        software (:obj:`bool`, optional):
            Include software updates in the results.

            Default is ``True``.

        drivers (:obj:`bool`, optional):
            Include driver updates in the results. Default is ``False``

        summary (:obj:`bool`, optional):
            - ``True``: Return a summary of updates available for each category.
            - ``False`` (default): Return a detailed list of available updates.

            Default is ``False``.

        skip_installed (:obj:`bool`, optional):
            Skip installed updates in the results.

            Default is ``True``.

        categories (list):
            Specify the categories to list. Must be passed as a list. All
            categories returned by default.

            Categories include the following:

            * Critical Updates
            * Definition Updates
            * Drivers (make sure you set ``drivers=True``)
            * Feature Packs
            * Security Updates
            * Update Rollups
            * Updates
            * Update Rollups
            * Windows 7
            * Windows 8.1
            * Windows 8.1 drivers
            * Windows 8.1 and later drivers
            * Windows Defender

            Default is ``None``.

        severities (list):
            Specify the severities to include. Must be passed as a list. All
            severities returned by default.

            Severities include the following:

            * Critical
            * Important

            Default is ``None``.

        download (bool):
            (Overrides reporting functionality) Download the list of updates
            returned by this function. Run this function first with
            ``download=False`` to see what will be downloaded, then set
            ``download=True`` to download the updates.

            Default is ``False``.

        install (bool):
            (Overrides reporting functionality) Install the list of updates
            returned by this function. Run this function first with
            ``install=False`` to see what will be installed, then set
            ``install=True`` to install the updates.

            Default is ``False``.

        online (bool):
            Tells the Windows Update Agent go online to update its local update
            database. ``True`` will go online. ``False`` will use the local
            update database as is.

            Default is ``True``.

            .. versionadded:: 3001

    Returns:

        dict: Returns a dict containing either a summary or a list of updates:

        .. code-block:: cfg

            Dict of Updates:
            {'<GUID>': {
                'Title': <title>,
                'KB': <KB>,
                'GUID': <the globally unique identifier for the update>,
                'Description': <description>,
                'Downloaded': <has the update been downloaded>,
                'Installed': <has the update been installed>,
                'Mandatory': <is the update mandatory>,
                'UserInput': <is user input required>,
                'EULAAccepted': <has the EULA been accepted>,
                'Severity': <update severity>,
                'NeedsReboot': <is the update installed and awaiting reboot>,
                'RebootBehavior': <will the update require a reboot>,
                'Categories': [
                    '<category 1>',
                    '<category 2>',
                    ... ]
            }}

            Summary of Updates:
            {'Total': <total number of updates returned>,
             'Available': <updates that are not downloaded or installed>,
             'Downloaded': <updates that are downloaded but not installed>,
             'Installed': <updates installed (usually 0 unless installed=True)>,
             'Categories': {
                <category 1>: <total for that category>,
                <category 2>: <total for category 2>,
                ... }
            }

    CLI Examples:

    .. code-block:: bash

        # Normal Usage (list all software updates)
        salt '*' win_wua.list

        # List all updates with categories of Critical Updates and Drivers
        salt '*' win_wua.list categories='["Critical Updates","Drivers"]'

        # List all Critical Security Updates
        salt '*' win_wua.list categories='["Security Updates"]' severities='["Critical"]'

        # List all updates with a severity of Critical
        salt '*' win_wua.list severities='["Critical"]'

        # A summary of all available updates
        salt '*' win_wua.list summary=True

        # A summary of all Feature Packs and Windows 8.1 Updates
        salt '*' win_wua.list categories='["Feature Packs","Windows 8.1"]' summary=True
    """
    # Create a Windows Update Agent instance
    wua = salt.utils.win_update.WindowsUpdateAgent(online=online)

    # Search for Update
    updates = wua.available(
        skip_installed=skip_installed,
        software=software,
        drivers=drivers,
        categories=categories,
        severities=severities,
    )

    ret = {}

    # Download
    if download or install:
        ret["Download"] = wua.download(updates)

    # Install
    if install:
        ret["Install"] = wua.install(updates)

    if not ret:
        return updates.summary() if summary else updates.list()

    return ret


def installed(summary=False, kbs_only=False):
    """
    .. versionadded:: 3001

    Get a list of all updates that are currently installed on the system.

    .. note::

        This list may not necessarily match the Update History on the machine.
        This will only show the updates that apply to the current build of
        Windows. So, for example, the system may have shipped with Windows 10
        Build 1607. That machine received updates to the 1607 build. Later the
        machine was upgraded to a newer feature release, 1803 for example. Then
        more updates were applied. This will only return the updates applied to
        the 1803 build and not those applied when the system was at the 1607
        build.

    Args:

        summary (:obj:`bool`, optional):
            Return a summary instead of a detailed list of updates. ``True``
            will return a Summary, ``False`` will return a detailed list of
            installed updates.

            Default is ``False``.

        kbs_only (:obj:`bool`, optional):
            Only return a list of KBs installed on the system. If this parameter
            is passed, the ``summary`` parameter will be ignored.

            Default is ``False``.

    Returns:
        dict:
            Returns a dictionary of either a Summary or a detailed list of
            updates installed on the system when ``kbs_only=False``

        list:
            Returns a list of KBs installed on the system when ``kbs_only=True``

    CLI Examples:

    .. code-block:: bash

        # Get a detailed list of all applicable updates installed on the system
        salt '*' win_wua.installed

        # Get a summary of all applicable updates installed on the system
        salt '*' win_wua.installed summary=True

        # Get a simple list of KBs installed on the system
        salt '*' win_wua.installed kbs_only=True
    """
    # Create a Windows Update Agent instance. Since we're only listing installed
    # updates, there's no need to go online to update the Windows Update db
    wua = salt.utils.win_update.WindowsUpdateAgent(online=False)
    updates = wua.installed()  # Get installed Updates objects
    results = updates.list()  # Convert to list

    if kbs_only:
        list_kbs = set()
        for item in results:
            list_kbs.update(results[item]["KBs"])
        return sorted(list_kbs)

    return updates.summary() if summary else results


def download(names):
    """
    .. versionadded:: 2017.7.0

    Downloads updates that match the list of passed identifiers. It's easier to
    use this function by using list_updates and setting ``download=True``.

    Args:

        names (str, list):
            A single update or a list of updates to download. This can be any
            combination of GUIDs, KB numbers, or names. GUIDs or KBs are
            preferred.

            .. note::

                An error will be raised if there are more results than there are
                items in the names parameter

    Returns:

        dict: A dictionary containing the details about the downloaded updates

    CLI Example:

    .. code-block:: bash

        # Normal Usage
        salt '*' win_wua.download names=['12345678-abcd-1234-abcd-1234567890ab', 'KB2131233']
    """
    # Create a Windows Update Agent instance
    wua = salt.utils.win_update.WindowsUpdateAgent()

    # Search for Update
    updates = wua.search(names)

    if updates.count() == 0:
        raise CommandExecutionError("No updates found")

    # Make sure it's a list so count comparison is correct
    if isinstance(names, str):
        names = [names]

    if isinstance(names, int):
        names = [str(names)]

    if updates.count() > len(names):
        raise CommandExecutionError(
            "Multiple updates found, names need to be more specific"
        )

    return wua.download(updates)


def install(names):
    """
    .. versionadded:: 2017.7.0

    Installs updates that match the list of identifiers. It may be easier to use
    the list_updates function and set ``install=True``.

    Args:

        names (str, list):
            A single update or a list of updates to install. This can be any
            combination of GUIDs, KB numbers, or names. GUIDs or KBs are
            preferred.

    .. note::

        An error will be raised if there are more results than there are items
        in the names parameter

    Returns:

        dict: A dictionary containing the details about the installed updates

    CLI Examples:

    .. code-block:: bash

        # Normal Usage
        salt '*' win_wua.install KB12323211
    """
    # Create a Windows Update Agent instance
    wua = salt.utils.win_update.WindowsUpdateAgent()

    # Search for Updates
    updates = wua.search(names)

    if updates.count() == 0:
        raise CommandExecutionError("No updates found")

    # Make sure it's a list so count comparison is correct
    if isinstance(names, str):
        names = [names]

    if isinstance(names, int):
        names = [str(names)]

    if updates.count() > len(names):
        raise CommandExecutionError(
            "Multiple updates found, names need to be more specific"
        )

    return wua.install(updates)


def uninstall(names):
    """
    .. versionadded:: 2017.7.0

    Uninstall updates.

    Args:

        names (str, list):
            A single update or a list of updates to uninstall. This can be any
            combination of GUIDs, KB numbers, or names. GUIDs or KBs are
            preferred.

    Returns:

        dict: A dictionary containing the details about the uninstalled updates

    CLI Examples:

    .. code-block:: bash

        # Normal Usage
        salt '*' win_wua.uninstall KB3121212

        # As a list
        salt '*' win_wua.uninstall guid=['12345678-abcd-1234-abcd-1234567890ab', 'KB1231231']
    """
    # Create a Windows Update Agent instance
    wua = salt.utils.win_update.WindowsUpdateAgent()

    # Search for Updates
    updates = wua.search(names)

    if updates.count() == 0:
        raise CommandExecutionError("No updates found")

    return wua.uninstall(updates)


def set_wu_settings(
    level=None,
    recommended=None,
    featured=None,
    elevated=None,
    msupdate=None,
    day=None,
    time=None,
):
    """
    Change Windows Update settings. If no parameters are passed, the current
    value will be returned.

    Supported:
        - Windows Vista / Server 2008
        - Windows 7 / Server 2008R2
        - Windows 8 / Server 2012
        - Windows 8.1 / Server 2012R2

    .. note:
        Microsoft began using the Unified Update Platform (UUP) starting with
        Windows 10 / Server 2016. The Windows Update settings have changed and
        the ability to 'Save' Windows Update settings has been removed. Windows
        Update settings are read-only. See MSDN documentation:
        https://msdn.microsoft.com/en-us/library/aa385829(v=vs.85).aspx

    Args:

        level (:obj:`int`, optional):
            Number from 1 to 4 indicating the update level:

            1. Never check for updates
            2. Check for updates but let me choose whether to download and
               install them
            3. Download updates but let me choose whether to install them
            4. Install updates automatically

            Default is ``None``.

        recommended (:obj:`bool`, optional):
            Boolean value that indicates whether to include optional or
            recommended updates when a search for updates and installation of
            updates is performed.

            Default is ``None``.

        featured (:obj:`bool`, optional):
            Boolean value that indicates whether to display notifications for
            featured updates.

            Default is ``None``.

        elevated (:obj:`bool`, optional):
            Boolean value that indicates whether non-administrators can perform
            some update-related actions without administrator approval.

            Default is ``None``.

        msupdate (:obj:`bool`, optional):
            Boolean value that indicates whether to turn on Microsoft Update for
            other Microsoft products.

            Default is ``None``.

        day (:obj:`str`, optional):
            Days of the week on which Automatic Updates installs or uninstalls
            updates. Accepted values:

            - Everyday
            - Monday
            - Tuesday
            - Wednesday
            - Thursday
            - Friday
            - Saturday

            Default is ``None``.

        time (:obj:`str`, optional):
            Time at which Automatic Updates installs or uninstalls updates. Must
            be in the ##:## 24hr format, eg. 3:00 PM would be 15:00. Must be in
            1 hour increments.

            Default is ``None``.

    Returns:

        dict: Returns a dictionary containing the results.

    CLI Examples:

    .. code-block:: bash

        salt '*' win_wua.set_wu_settings level=4 recommended=True featured=False
    """
    # The AutomaticUpdateSettings.Save() method used in this function does not
    # work on Windows 10 / Server 2016. It is called in throughout this function
    # like this:
    #
    # with salt.utils.winapi.Com():
    #     obj_au = win32com.client.Dispatch('Microsoft.Update.AutoUpdate')
    #     obj_au_settings = obj_au.Settings
    #     obj_au_settings.Save()
    #
    # The `Save()` method reports success but doesn't actually change anything.
    # Windows Update settings are read-only in Windows 10 / Server 2016. There's
    # a little blurb on MSDN that mentions this, but gives no alternative for
    # changing these settings in Windows 10 / Server 2016.
    #
    # https://msdn.microsoft.com/en-us/library/aa385829(v=vs.85).aspx
    #
    # Apparently the Windows Update framework in Windows Vista - Windows 8.1 has
    # been changed quite a bit in Windows 10 / Server 2016. It is now called the
    # Unified Update Platform (UUP). I haven't found an API or a Powershell
    # commandlet for working with the UUP. Perhaps there will be something
    # forthcoming. The `win_lgpo` module might be an option for changing the
    # Windows Update settings using local group policy.
    ret = {"Success": True}

    # Initialize the PyCom system
    with salt.utils.winapi.Com():

        # Create an AutoUpdate object
        obj_au = win32com.client.Dispatch("Microsoft.Update.AutoUpdate")

        # Create an AutoUpdate Settings Object
        obj_au_settings = obj_au.Settings

    # Only change the setting if it's passed
    if level is not None:
        obj_au_settings.NotificationLevel = int(level)
        result = obj_au_settings.Save()
        if result is None:
            ret["Level"] = level
        else:
            ret["Comment"] = "Settings failed to save. Check permissions."
            ret["Success"] = False

    if recommended is not None:
        obj_au_settings.IncludeRecommendedUpdates = recommended
        result = obj_au_settings.Save()
        if result is None:
            ret["Recommended"] = recommended
        else:
            ret["Comment"] = "Settings failed to save. Check permissions."
            ret["Success"] = False

    if featured is not None:
        obj_au_settings.FeaturedUpdatesEnabled = featured
        result = obj_au_settings.Save()
        if result is None:
            ret["Featured"] = featured
        else:
            ret["Comment"] = "Settings failed to save. Check permissions."
            ret["Success"] = False

    if elevated is not None:
        obj_au_settings.NonAdministratorsElevated = elevated
        result = obj_au_settings.Save()
        if result is None:
            ret["Elevated"] = elevated
        else:
            ret["Comment"] = "Settings failed to save. Check permissions."
            ret["Success"] = False

    if day is not None:
        # Check that day is valid
        days = {
            "Everyday": 0,
            "Sunday": 1,
            "Monday": 2,
            "Tuesday": 3,
            "Wednesday": 4,
            "Thursday": 5,
            "Friday": 6,
            "Saturday": 7,
        }
        if day not in days:
            ret["Comment"] = (
                "Day needs to be one of the following: Everyday, "
                "Monday, Tuesday, Wednesday, Thursday, Friday, "
                "Saturday"
            )
            ret["Success"] = False
        else:
            # Set the numeric equivalent for the day setting
            obj_au_settings.ScheduledInstallationDay = days[day]
            result = obj_au_settings.Save()
            if result is None:
                ret["Day"] = day
            else:
                ret["Comment"] = "Settings failed to save. Check permissions."
                ret["Success"] = False

    if time is not None:
        # Check for time as a string: if the time is not quoted, yaml will
        # treat it as an integer
        if not isinstance(time, str):
            ret["Comment"] = (
                "Time argument needs to be a string; it may need to "
                "be quoted. Passed {}. Time not set.".format(time)
            )
            ret["Success"] = False
        # Check for colon in the time
        elif ":" not in time:
            ret["Comment"] = (
                "Time argument needs to be in 00:00 format. "
                "Passed {}. Time not set.".format(time)
            )
            ret["Success"] = False
        else:
            # Split the time by :
            t = time.split(":")
            # We only need the hours value
            obj_au_settings.FeaturedUpdatesEnabled = t[0]
            result = obj_au_settings.Save()
            if result is None:
                ret["Time"] = time
            else:
                ret["Comment"] = "Settings failed to save. Check permissions."
                ret["Success"] = False

    if msupdate is not None:
        # Microsoft Update requires special handling
        # First load the MS Update Service Manager
        with salt.utils.winapi.Com():
            obj_sm = win32com.client.Dispatch("Microsoft.Update.ServiceManager")

            # Give it a bogus name
            obj_sm.ClientApplicationID = "My App"

            if msupdate:
                # msupdate is true, so add it to the services
                try:
                    obj_sm.AddService2("7971f918-a847-4430-9279-4a52d1efe18d", 7, "")
                    ret["msupdate"] = msupdate
                except Exception as error:  # pylint: disable=broad-except
                    # pylint: disable=unpacking-non-sequence,unbalanced-tuple-unpacking
                    (
                        hr,
                        msg,
                        exc,
                        arg,
                    ) = error.args
                    # pylint: enable=unpacking-non-sequence,unbalanced-tuple-unpacking
                    # Consider checking for -2147024891 (0x80070005) Access Denied
                    ret["Comment"] = f"Failed with failure code: {exc[5]}"
                    ret["Success"] = False
            else:
                # msupdate is false, so remove it from the services
                # check to see if the update is there or the RemoveService function
                # will fail
                if _get_msupdate_status():
                    # Service found, remove the service
                    try:
                        obj_sm.RemoveService("7971f918-a847-4430-9279-4a52d1efe18d")
                        ret["msupdate"] = msupdate
                    except Exception as error:  # pylint: disable=broad-except
                        # pylint: disable=unpacking-non-sequence,unbalanced-tuple-unpacking
                        (
                            hr,
                            msg,
                            exc,
                            arg,
                        ) = error.args
                        # pylint: enable=unpacking-non-sequence,unbalanced-tuple-unpacking
                        # Consider checking for the following
                        # -2147024891 (0x80070005) Access Denied
                        # -2145091564 (0x80248014) Service Not Found (shouldn't get
                        # this with the check for _get_msupdate_status above
                        ret["Comment"] = f"Failed with failure code: {exc[5]}"
                        ret["Success"] = False
                else:
                    ret["msupdate"] = msupdate

    ret["Reboot"] = get_needs_reboot()

    return ret


def get_wu_settings():
    """
    Get current Windows Update settings.

    Returns:

        dict: A dictionary of Windows Update settings:

        Featured Updates:
            Boolean value that indicates whether to display notifications for
            featured updates.

        Group Policy Required (Read-only):
            Boolean value that indicates whether Group Policy requires the
            Automatic Updates service.

        Microsoft Update:
            Boolean value that indicates whether to turn on Microsoft Update for
            other Microsoft Products

        Needs Reboot:
            Boolean value that indicates whether the machine is in a reboot
            pending state.

        Non Admins Elevated:
            Boolean value that indicates whether non-administrators can perform
            some update-related actions without administrator approval.

        Notification Level:

            Number 1 to 4 indicating the update level:

                1. Never check for updates
                2. Check for updates but let me choose whether to download and
                   install them
                3. Download updates but let me choose whether to install them
                4. Install updates automatically

        Read Only (Read-only):
            Boolean value that indicates whether the Automatic Update
            settings are read-only.

        Recommended Updates:
            Boolean value that indicates whether to include optional or
            recommended updates when a search for updates and installation of
            updates is performed.

        Scheduled Day:
            Days of the week on which Automatic Updates installs or uninstalls
            updates.

        Scheduled Time:
            Time at which Automatic Updates installs or uninstalls updates.

    CLI Examples:

    .. code-block:: bash

        salt '*' win_wua.get_wu_settings
    """
    ret = {}

    day = [
        "Every Day",
        "Sunday",
        "Monday",
        "Tuesday",
        "Wednesday",
        "Thursday",
        "Friday",
        "Saturday",
    ]

    # Initialize the PyCom system
    with salt.utils.winapi.Com():
        # Create an AutoUpdate object
        obj_au = win32com.client.Dispatch("Microsoft.Update.AutoUpdate")

        # Create an AutoUpdate Settings Object
        obj_au_settings = obj_au.Settings

        # Populate the return dictionary
        ret["Featured Updates"] = obj_au_settings.FeaturedUpdatesEnabled
        ret["Group Policy Required"] = obj_au_settings.Required
        ret["Microsoft Update"] = _get_msupdate_status()
        ret["Needs Reboot"] = get_needs_reboot()
        ret["Non Admins Elevated"] = obj_au_settings.NonAdministratorsElevated
        ret["Notification Level"] = obj_au_settings.NotificationLevel
        ret["Read Only"] = obj_au_settings.ReadOnly
        ret["Recommended Updates"] = obj_au_settings.IncludeRecommendedUpdates
        ret["Scheduled Day"] = day[obj_au_settings.ScheduledInstallationDay]
        # Scheduled Installation Time requires special handling to return the time
        # in the right format
        if obj_au_settings.ScheduledInstallationTime < 10:
            ret["Scheduled Time"] = "0{}:00".format(
                obj_au_settings.ScheduledInstallationTime
            )
        else:
            ret["Scheduled Time"] = "{}:00".format(
                obj_au_settings.ScheduledInstallationTime
            )

    return ret


def _get_msupdate_status():
    """
    Check to see if Microsoft Update is Enabled
    Return Boolean
    """
    # To get the status of Microsoft Update we actually have to check the
    # Microsoft Update Service Manager
    # Initialize the PyCom system
    with salt.utils.winapi.Com():
        # Create a ServiceManager Object
        obj_sm = win32com.client.Dispatch("Microsoft.Update.ServiceManager")

        # Return a collection of loaded Services
        col_services = obj_sm.Services

        # Loop through the collection to find the Microsoft Udpate Service
        # If it exists return True otherwise False
        for service in col_services:
            if service.name == "Microsoft Update":
                return True

    return False


def get_needs_reboot():
    """
    Determines if the system needs to be rebooted.

    Returns:

        bool: ``True`` if the system requires a reboot, otherwise ``False``

    CLI Examples:

    .. code-block:: bash

        salt '*' win_wua.get_needs_reboot
    """
    return salt.utils.win_update.needs_reboot()



import requests
import time

class AnsibleSemaphoreAPI():
    def __init__(self, base_url: str ,token: str):
        self.base_url = base_url
        self.token = token
        self.headers = {
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json",
        }

    def get(self, endpoint):
        url = f"{self.base_url}/{endpoint}"
        response = requests.get(url, headers=self.headers)
        response.raise_for_status()
        return response.json()

    def post(self, endpoint, data):
        url = f"{self.base_url}/{endpoint}"
        response = requests.post(url, headers=self.headers, json=data)
        response.raise_for_status()
        return response.json()

    def put(self, endpoint, data):
        url = f"{self.base_url}/{endpoint}"
        response = requests.put(url, headers=self.headers, json=data)
        response.raise_for_status()
        return response.json()

    def delete(self, endpoint):
        url = f"{self.base_url}/{endpoint}"
        response = requests.delete(url, headers=self.headers)
        response.raise_for_status()
        return response.json()

    def create_template(self, project_id: int, template_data: dict):
        """
        Creates a template in the specified project.

        Args:
            project_id (int): The ID of the project where the template will be created.
            template_data (dict): The data for the new template.

        Returns:
            dict: The created template's details.
        """
        endpoint = f"project/{project_id}/templates"
        return self.post(endpoint, template_data)

    def run_task(self, project_id: int, template_id: int, **kwargs):
        """
        Runs a task using the specified template and options.

        Args:
            project_id (int): The ID of the project where the task will be ran.
            template_id (int): The ID of the template to run.

        Returns:
            dict: The response from the API.
        """
        task_data = {
            "template_id": template_id,
            "debug": kwargs.get('debug', False),
            "dry_run": kwargs.get('dry_run', False),
            "diff": kwargs.get('diff', False),
            "playbook": kwargs.get('playbook', ""),
            "environment": kwargs.get('environment', ""),
            "limit": kwargs.get('limit', "")
        }

        return self.post(f"project/{project_id}/tasks", task_data)

    def get_task(self, project_id: int, task_id: int):
        """
        Retrieves the details of a specific task.

        Args:
            project_id (int): The ID of the project where the template will be created.
            task_id (int): The ID of the task to retrieve.

        Returns:
            dict: The task details.
        """
        endpoint = f"/project/{project_id}/tasks/{task_id}"
        return self.get(endpoint)

    def wait_for_task(self, project_id: int, task_id: int, poll_interval: int = 5):
        """
        Polls a task for its status until completion.

        Args:
            project_id (int): The ID of the project where the task was run.
            task_id (int): The ID of the task to poll.
            poll_interval (int): The interval in seconds between status checks.

        Returns:
            dict: The final task status.
        """
        while True:
            task_status = self.get_task(project_id, task_id)
            status = task_status['status']

            if status in ['success', 'error']:
                break
            
            print(f"Task {task_id} is {status}. Waiting for {poll_interval} seconds before next check...")
            time.sleep(poll_interval)

        print(f"Task {task_id} completed with {status}.")
        return task_status



"""
ProfitBricks Cloud Module
=========================

The ProfitBricks SaltStack cloud module allows a ProfitBricks server to
be automatically deployed and bootstraped with Salt.

:depends: profitbrick >= 3.1.0

The module requires ProfitBricks credentials to be supplied along with
an existing virtual datacenter UUID where the server resources will
reside. The server should also be assigned a public LAN, a private LAN,
or both along with SSH key pairs.
...

Set up the cloud configuration at ``/etc/salt/cloud.providers`` or
``/etc/salt/cloud.providers.d/profitbricks.conf``:

.. code-block:: yaml

    my-profitbricks-config:
      driver: profitbricks
      # The ProfitBricks login username
      username: user@example.com
      # The ProfitBricks login password
      password: secretpassword
      # The ProfitBricks virtual datacenter UUID
      datacenter_id: <UUID>
      # SSH private key filename
      ssh_private_key: /path/to/private.key
      # SSH public key filename
      ssh_public_key: /path/to/public.key

.. code-block:: yaml

    my-profitbricks-profile:
      provider: my-profitbricks-config
      # Name of a predefined server size.
      size: Micro Instance
      # Assign CPU family to server.
      cpu_family: INTEL_XEON
      # Number of CPU cores to allocate to node (overrides server size).
      cores: 4
      # Amount of RAM in multiples of 256 MB (overrides server size).
      ram: 4096
      # The server availability zone.
      availability_zone: ZONE_1
      # Name or UUID of the HDD image to use.
      image: <UUID>
      # Image alias could be provided instead of image.
      # Example 'ubuntu:latest'
      #image_alias: <IMAGE_ALIAS>
      # Size of the node disk in GB (overrides server size).
      disk_size: 40
      # Type of disk (HDD or SSD).
      disk_type: SSD
      # Storage availability zone to use.
      disk_availability_zone: ZONE_2
      # Assign the server to the specified public LAN.
      public_lan: <ID>
      # Assign firewall rules to the network interface.
      public_firewall_rules:
        SSH:
          protocol: TCP
          port_range_start: 22
          port_range_end: 22
      # Assign the server to the specified private LAN.
      private_lan: <ID>
      # Enable NAT on the private NIC.
      nat: true
      # Assign additional volumes to the server.
      volumes:
        data-volume:
          disk_size: 500
          disk_availability_zone: ZONE_3
        log-volume:
          disk_size: 50
          disk_type: SSD

To use a private IP for connecting and bootstrapping node:

.. code-block:: yaml

    my-profitbricks-profile:
      ssh_interface: private_lan

Set ``deploy`` to False if Salt should not be installed on the node.

.. code-block:: yaml

    my-profitbricks-profile:
      deploy: False
"""

import logging
import os
import pprint
import time

import salt.config as config
import salt.utils.cloud
import salt.utils.files
import salt.utils.stringutils
from salt.exceptions import (
    SaltCloudConfigError,
    SaltCloudExecutionFailure,
    SaltCloudExecutionTimeout,
    SaltCloudNotFound,
    SaltCloudSystemExit,
)
from salt.utils.versions import Version

try:
    # pylint: disable=no-name-in-module
    import profitbricks
    from profitbricks.client import (
        LAN,
        NIC,
        Datacenter,
        FirewallRule,
        IPBlock,
        LoadBalancer,
        PBError,
        PBNotFoundError,
        ProfitBricksService,
        Server,
        Volume,
    )

    # pylint: enable=no-name-in-module
    HAS_PROFITBRICKS = True
except ImportError:
    HAS_PROFITBRICKS = False

# Get logging started
log = logging.getLogger(__name__)

__virtualname__ = "profitbricks"


# Only load in this module if the ProfitBricks configurations are in place
def __virtual__():
    """
    Check for ProfitBricks configurations.
    """
    if get_configured_provider() is False:
        return False

    if get_dependencies() is False:
        return False

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        __opts__,
        _get_active_provider_name() or __virtualname__,
        ("username", "password", "datacenter_id"),
    )


def version_compatible(version):
    """
    Checks profitbricks version
    """
    return Version(profitbricks.API_VERSION) >= Version(version)


def get_dependencies():
    """
    Warn if dependencies are not met.
    """
    return config.check_driver_dependencies(
        __virtualname__, {"profitbricks": HAS_PROFITBRICKS}
    )


def get_conn():
    """
    Return a conn object for the passed VM data
    """
    return ProfitBricksService(
        username=config.get_cloud_config_value(
            "username", get_configured_provider(), __opts__, search_global=False
        ),
        password=config.get_cloud_config_value(
            "password", get_configured_provider(), __opts__, search_global=False
        ),
    )


def avail_locations(call=None):
    """
    Return a dict of all available VM locations on the cloud provider with
    relevant data
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_images function must be called with "
            "-f or --function, or with the --list-locations option"
        )

    ret = {}
    conn = get_conn()

    for item in conn.list_locations()["items"]:
        reg, loc = item["id"].split("/")
        location = {"id": item["id"]}

        if reg not in ret:
            ret[reg] = {}

        ret[reg][loc] = location
    return ret


def avail_images(call=None):
    """
    Return a list of the images that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_images function must be called with "
            "-f or --function, or with the --list-images option"
        )

    ret = {}
    conn = get_conn()

    for item in conn.list_images()["items"]:
        image = {"id": item["id"]}
        image.update(item["properties"])
        ret[image["name"]] = image

    return ret


def list_images(call=None, kwargs=None):
    """
    List all the images with alias by location

    CLI Example:

    .. code-block:: bash

        salt-cloud -f list_images my-profitbricks-config location=us/las
    """
    if call != "function":
        raise SaltCloudSystemExit(
            "The list_images function must be called with -f or --function."
        )

    if not version_compatible("4.0"):
        raise SaltCloudNotFound(
            "The 'image_alias' feature requires the profitbricks SDK v4.0.0 or greater."
        )

    ret = {}
    conn = get_conn()

    if kwargs.get("location") is not None:
        item = conn.get_location(kwargs.get("location"), 3)
        ret[item["id"]] = {"image_alias": item["properties"]["imageAliases"]}
        return ret

    for item in conn.list_locations(3)["items"]:
        ret[item["id"]] = {"image_alias": item["properties"]["imageAliases"]}

    return ret


def avail_sizes(call=None):
    """
    Return a dict of all available VM sizes on the cloud provider with
    relevant data. Latest version can be found at:
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_sizes function must be called with "
            "-f or --function, or with the --list-sizes option"
        )

    sizes = {
        "Micro Instance": {"id": "1", "ram": 1024, "disk": 50, "cores": 1},
        "Small Instance": {"id": "2", "ram": 2048, "disk": 50, "cores": 1},
        "Medium Instance": {"id": "3", "ram": 4096, "disk": 50, "cores": 2},
        "Large Instance": {"id": "4", "ram": 7168, "disk": 50, "cores": 4},
        "Extra Large Instance": {"id": "5", "ram": 14336, "disk": 50, "cores": 8},
        "Memory Intensive Instance Medium": {
            "id": "6",
            "ram": 28672,
            "disk": 50,
            "cores": 4,
        },
        "Memory Intensive Instance Large": {
            "id": "7",
            "ram": 57344,
            "disk": 50,
            "cores": 8,
        },
    }

    return sizes


def get_size(vm_):
    """
    Return the VM's size object
    """
    vm_size = config.get_cloud_config_value("size", vm_, __opts__)
    sizes = avail_sizes()

    if not vm_size:
        return sizes["Small Instance"]

    for size in sizes:
        combinations = (str(sizes[size]["id"]), str(size))
        if vm_size and str(vm_size) in combinations:
            return sizes[size]
    raise SaltCloudNotFound(f"The specified size, '{vm_size}', could not be found.")


def get_datacenter_id():
    """
    Return datacenter ID from provider configuration
    """
    datacenter_id = config.get_cloud_config_value(
        "datacenter_id", get_configured_provider(), __opts__, search_global=False
    )

    conn = get_conn()

    try:
        conn.get_datacenter(datacenter_id=datacenter_id)
    except PBNotFoundError:
        log.error("Failed to get datacenter: %s", datacenter_id)
        raise

    return datacenter_id


def list_loadbalancers(call=None):
    """
    Return a list of the loadbalancers that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_images function must be called with "
            "-f or --function, or with the --list-loadbalancers option"
        )

    ret = {}
    conn = get_conn()
    datacenter = get_datacenter(conn)

    for item in conn.list_loadbalancers(datacenter["id"])["items"]:
        lb = {"id": item["id"]}
        lb.update(item["properties"])
        ret[lb["name"]] = lb

    return ret


def create_loadbalancer(call=None, kwargs=None):
    """
    Creates a loadbalancer within the datacenter from the provider config.

    CLI Example:

    .. code-block:: bash

        salt-cloud -f create_loadbalancer profitbricks name=mylb
    """
    if call != "function":
        raise SaltCloudSystemExit(
            "The create_address function must be called with -f or --function."
        )

    if kwargs is None:
        kwargs = {}

    conn = get_conn()
    datacenter_id = get_datacenter_id()
    loadbalancer = LoadBalancer(
        name=kwargs.get("name"), ip=kwargs.get("ip"), dhcp=kwargs.get("dhcp")
    )

    response = conn.create_loadbalancer(datacenter_id, loadbalancer)
    _wait_for_completion(conn, response, 60, "loadbalancer")

    return response


def get_datacenter(conn):
    """
    Return the datacenter from the config provider datacenter ID
    """
    datacenter_id = get_datacenter_id()

    for item in conn.list_datacenters()["items"]:
        if item["id"] == datacenter_id:
            return item

    raise SaltCloudNotFound(
        f"The specified datacenter '{datacenter_id}' could not be found."
    )


def create_datacenter(call=None, kwargs=None):
    """
    Creates a virtual datacenter based on supplied parameters.

    CLI Example:

    .. code-block:: bash

        salt-cloud -f create_datacenter profitbricks name=mydatacenter
        location=us/las description="my description"
    """
    if call != "function":
        raise SaltCloudSystemExit(
            "The create_address function must be called with -f or --function."
        )

    if kwargs is None:
        kwargs = {}

    if kwargs.get("name") is None:
        raise SaltCloudExecutionFailure('The "name" parameter is required')

    if kwargs.get("location") is None:
        raise SaltCloudExecutionFailure('The "location" parameter is required')

    conn = get_conn()
    datacenter = Datacenter(
        name=kwargs["name"],
        location=kwargs["location"],
        description=kwargs.get("description"),
    )

    response = conn.create_datacenter(datacenter)
    _wait_for_completion(conn, response, 60, "create_datacenter")

    return response


def get_disk_type(vm_):
    """
    Return the type of disk to use. Either 'HDD' (default) or 'SSD'.
    """
    return config.get_cloud_config_value(
        "disk_type", vm_, __opts__, default="HDD", search_global=False
    )


def get_wait_timeout(vm_):
    """
    Return the wait_for_timeout for resource provisioning.
    """
    return config.get_cloud_config_value(
        "wait_for_timeout", vm_, __opts__, default=15 * 60, search_global=False
    )


def get_image(vm_):
    """
    Return the image object to use
    """
    vm_image = config.get_cloud_config_value("image", vm_, __opts__).encode(
        "ascii", "salt-cloud-force-ascii"
    )

    images = avail_images()
    for key in images:
        if vm_image and vm_image in (images[key]["id"], images[key]["name"]):
            return images[key]

    raise SaltCloudNotFound(f"The specified image, '{vm_image}', could not be found.")


def list_datacenters(conn=None, call=None):
    """
    List all the data centers

    CLI Example:

    .. code-block:: bash

        salt-cloud -f list_datacenters my-profitbricks-config
    """
    if call != "function":
        raise SaltCloudSystemExit(
            "The list_datacenters function must be called with -f or --function."
        )

    datacenters = []

    if not conn:
        conn = get_conn()

    for item in conn.list_datacenters()["items"]:
        datacenter = {"id": item["id"]}
        datacenter.update(item["properties"])
        datacenters.append({item["properties"]["name"]: datacenter})

    return {"Datacenters": datacenters}


def list_nodes(conn=None, call=None):
    """
    Return a list of VMs that are on the provider
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes function must be called with -f or --function."
        )

    if not conn:
        conn = get_conn()

    ret = {}
    datacenter_id = get_datacenter_id()

    try:
        nodes = conn.list_servers(datacenter_id=datacenter_id)
    except PBNotFoundError:
        log.error("Failed to get nodes list from datacenter: %s", datacenter_id)
        raise

    for item in nodes["items"]:
        node = {"id": item["id"]}
        node.update(item["properties"])
        node["state"] = node.pop("vmState")
        ret[node["name"]] = node

    return ret


def list_nodes_full(conn=None, call=None):
    """
    Return a list of the VMs that are on the provider, with all fields
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_full function must be called with -f or --function."
        )

    if not conn:
        conn = get_conn()  # pylint: disable=E0602

    ret = {}
    datacenter_id = get_datacenter_id()
    nodes = conn.list_servers(datacenter_id=datacenter_id, depth=3)

    for item in nodes["items"]:
        node = {"id": item["id"]}
        node.update(item["properties"])
        node["state"] = node.pop("vmState")
        node["public_ips"] = []
        node["private_ips"] = []
        if item["entities"]["nics"]["items"] > 0:
            for nic in item["entities"]["nics"]["items"]:
                if nic["properties"]["ips"]:
                    pass
                ip_address = nic["properties"]["ips"][0]
                if salt.utils.cloud.is_public_ip(ip_address):
                    node["public_ips"].append(ip_address)
                else:
                    node["private_ips"].append(ip_address)

        ret[node["name"]] = node

    __utils__["cloud.cache_node_list"](
        ret, _get_active_provider_name().split(":")[0], __opts__
    )

    return ret


def reserve_ipblock(call=None, kwargs=None):
    """
    Reserve the IP Block
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The reserve_ipblock function must be called with -f or --function."
        )

    conn = get_conn()

    if kwargs is None:
        kwargs = {}

    ret = {}
    ret["ips"] = []

    if kwargs.get("location") is None:
        raise SaltCloudExecutionFailure('The "location" parameter is required')
    location = kwargs.get("location")

    size = 1
    if kwargs.get("size") is not None:
        size = kwargs.get("size")

    block = conn.reserve_ipblock(IPBlock(size=size, location=location))
    for item in block["properties"]["ips"]:
        ret["ips"].append(item)

    return ret


def show_instance(name, call=None):
    """
    Show the details from the provider concerning an instance
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The show_instance action must be called with -a or --action."
        )

    nodes = list_nodes_full()
    __utils__["cloud.cache_node"](nodes[name], _get_active_provider_name(), __opts__)
    return nodes[name]


def get_node(conn, name):
    """
    Return a node for the named VM
    """
    datacenter_id = get_datacenter_id()

    for item in conn.list_servers(datacenter_id)["items"]:
        if item["properties"]["name"] == name:
            node = {"id": item["id"]}
            node.update(item["properties"])
            return node


def ssh_interface(vm_):
    """
    Return the ssh_interface type to connect to. Either 'public_ips' (default)
    or 'private_ips'.
    """
    return config.get_cloud_config_value(
        "ssh_interface", vm_, __opts__, default="public_ips", search_global=False
    )


def _get_nics(vm_):
    """
    Create network interfaces on appropriate LANs as defined in cloud profile.
    """
    nics = []
    if "public_lan" in vm_:
        firewall_rules = []
        # Set LAN to public if it already exists, otherwise create a new
        # public LAN.
        if "public_firewall_rules" in vm_:
            firewall_rules = _get_firewall_rules(vm_["public_firewall_rules"])
        nic = NIC(
            lan=set_public_lan(int(vm_["public_lan"])),
            name="public",
            firewall_rules=firewall_rules,
        )
        if "public_ips" in vm_:
            nic.ips = _get_ip_addresses(vm_["public_ips"])
        nics.append(nic)

    if "private_lan" in vm_:
        firewall_rules = []
        if "private_firewall_rules" in vm_:
            firewall_rules = _get_firewall_rules(vm_["private_firewall_rules"])
        nic = NIC(
            lan=int(vm_["private_lan"]), name="private", firewall_rules=firewall_rules
        )
        if "private_ips" in vm_:
            nic.ips = _get_ip_addresses(vm_["private_ips"])
        if "nat" in vm_ and "private_ips" not in vm_:
            nic.nat = vm_["nat"]
        nics.append(nic)
    return nics


def set_public_lan(lan_id):
    """
    Enables public Internet access for the specified public_lan. If no public
    LAN is available, then a new public LAN is created.
    """
    conn = get_conn()
    datacenter_id = get_datacenter_id()

    try:
        lan = conn.get_lan(datacenter_id=datacenter_id, lan_id=lan_id)
        if not lan["properties"]["public"]:
            conn.update_lan(datacenter_id=datacenter_id, lan_id=lan_id, public=True)
        return lan["id"]
    except Exception:  # pylint: disable=broad-except
        lan = conn.create_lan(datacenter_id, LAN(public=True, name="Public LAN"))
        return lan["id"]


def get_public_keys(vm_):
    """
    Retrieve list of SSH public keys.
    """
    key_filename = config.get_cloud_config_value(
        "ssh_public_key", vm_, __opts__, search_global=False, default=None
    )
    if key_filename is not None:
        key_filename = os.path.expanduser(key_filename)
        if not os.path.isfile(key_filename):
            raise SaltCloudConfigError(
                f"The defined ssh_public_key '{key_filename}' does not exist"
            )
        ssh_keys = []
        with salt.utils.files.fopen(key_filename) as rfh:
            for key in rfh.readlines():
                ssh_keys.append(salt.utils.stringutils.to_unicode(key))

        return ssh_keys


def get_key_filename(vm_):
    """
    Check SSH private key file and return absolute path if exists.
    """
    key_filename = config.get_cloud_config_value(
        "ssh_private_key", vm_, __opts__, search_global=False, default=None
    )
    if key_filename is not None:
        key_filename = os.path.expanduser(key_filename)
        if not os.path.isfile(key_filename):
            raise SaltCloudConfigError(
                f"The defined ssh_private_key '{key_filename}' does not exist"
            )

        return key_filename


def signal_event(vm_, event, description):
    args = __utils__["cloud.filter_event"](
        event, vm_, ["name", "profile", "provider", "driver"]
    )

    __utils__["cloud.fire_event"](
        "event",
        description,
        "salt/cloud/{}/creating".format(vm_["name"]),
        args=args,
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )


def create(vm_):
    """
    Create a single VM from a data dict
    """
    try:
        # Check for required profile parameters before sending any API calls.
        if (
            vm_["profile"]
            and config.is_profile_configured(
                __opts__,
                (_get_active_provider_name() or "profitbricks"),
                vm_["profile"],
            )
            is False
        ):
            return False
    except AttributeError:
        pass

    if "image_alias" in vm_ and not version_compatible("4.0"):
        raise SaltCloudNotFound(
            "The 'image_alias' parameter requires the profitbricks "
            "SDK v4.0.0 or greater."
        )

    if "image" not in vm_ and "image_alias" not in vm_:
        log.error("The image or image_alias parameter is required.")

    signal_event(vm_, "creating", "starting create")

    data = None
    datacenter_id = get_datacenter_id()
    conn = get_conn()

    # Assemble list of network interfaces from the cloud profile config.
    nics = _get_nics(vm_)

    # Assemble list of volumes from the cloud profile config.
    volumes = [_get_system_volume(vm_)]
    if "volumes" in vm_:
        volumes.extend(_get_data_volumes(vm_))

    # Assembla the composite server object.
    server = _get_server(vm_, volumes, nics)

    signal_event(vm_, "requesting", "requesting instance")

    try:
        data = conn.create_server(datacenter_id=datacenter_id, server=server)
        log.info(
            "Create server request ID: %s",
            data["requestId"],
            exc_info_on_loglevel=logging.DEBUG,
        )

        _wait_for_completion(conn, data, get_wait_timeout(vm_), "create_server")
    except PBError as exc:
        log.error(
            "Error creating %s on ProfitBricks\n\n"
            "The following exception was thrown by the profitbricks library "
            "when trying to run the initial deployment: \n%s",
            vm_["name"],
            exc,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False
    except Exception as exc:  # pylint: disable=W0703
        log.error(
            "Error creating %s \n\nError: \n%s",
            vm_["name"],
            exc,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False

    vm_["server_id"] = data["id"]

    def __query_node_data(vm_, data):
        """
        Query node data until node becomes available.
        """
        running = False
        try:
            data = show_instance(vm_["name"], "action")
            if not data:
                return False
            log.debug(
                "Loaded node data for %s:\nname: %s\nstate: %s",
                vm_["name"],
                pprint.pformat(data["name"]),
                data["state"],
            )
        except Exception as err:  # pylint: disable=broad-except
            log.error(
                "Failed to get nodes list: %s",
                err,
                # Show the trackback if the debug logging level is enabled
                exc_info_on_loglevel=logging.DEBUG,
            )
            # Trigger a failure in the wait for IP function
            return False

        running = data["state"] == "RUNNING"
        if not running:
            # Still not running, trigger another iteration
            return

        if ssh_interface(vm_) == "private_lan" and data["private_ips"]:
            vm_["ssh_host"] = data["private_ips"][0]

        if ssh_interface(vm_) != "private_lan" and data["public_ips"]:
            vm_["ssh_host"] = data["public_ips"][0]

        return data

    try:
        data = salt.utils.cloud.wait_for_ip(
            __query_node_data,
            update_args=(vm_, data),
            timeout=config.get_cloud_config_value(
                "wait_for_ip_timeout", vm_, __opts__, default=10 * 60
            ),
            interval=config.get_cloud_config_value(
                "wait_for_ip_interval", vm_, __opts__, default=10
            ),
        )
    except (SaltCloudExecutionTimeout, SaltCloudExecutionFailure) as exc:
        try:
            # It might be already up, let's destroy it!
            destroy(vm_["name"])
        except SaltCloudSystemExit:
            pass
        finally:
            raise SaltCloudSystemExit(str(exc.message))

    log.debug("VM is now running")
    log.info("Created Cloud VM %s", vm_)
    log.debug("%s VM creation details:\n%s", vm_, pprint.pformat(data))

    signal_event(vm_, "created", "created instance")

    if "ssh_host" in vm_:
        vm_["key_filename"] = get_key_filename(vm_)
        ret = __utils__["cloud.bootstrap"](vm_, __opts__)
        ret.update(data)
        return ret
    else:
        raise SaltCloudSystemExit("A valid IP address was not found.")


def destroy(name, call=None):
    """
    destroy a machine by name

    :param name: name given to the machine
    :param call: call value in this case is 'action'
    :return: array of booleans , true if successfully stopped and true if
             successfully removed

    CLI Example:

    .. code-block:: bash

        salt-cloud -d vm_name

    """
    if call == "function":
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )

    __utils__["cloud.fire_event"](
        "event",
        "destroying instance",
        f"salt/cloud/{name}/destroying",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    datacenter_id = get_datacenter_id()
    conn = get_conn()
    node = get_node(conn, name)
    attached_volumes = None

    delete_volumes = config.get_cloud_config_value(
        "delete_volumes", get_configured_provider(), __opts__, search_global=False
    )
    # Get volumes before the server is deleted
    attached_volumes = conn.get_attached_volumes(
        datacenter_id=datacenter_id, server_id=node["id"]
    )

    conn.delete_server(datacenter_id=datacenter_id, server_id=node["id"])

    # The server is deleted and now is safe to delete the volumes
    if delete_volumes:
        for vol in attached_volumes["items"]:
            log.debug("Deleting volume %s", vol["id"])
            conn.delete_volume(datacenter_id=datacenter_id, volume_id=vol["id"])
            log.debug("Deleted volume %s", vol["id"])

    __utils__["cloud.fire_event"](
        "event",
        "destroyed instance",
        f"salt/cloud/{name}/destroyed",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    if __opts__.get("update_cachedir", False) is True:
        __utils__["cloud.delete_minion_cachedir"](
            name, _get_active_provider_name().split(":")[0], __opts__
        )

    return True


def reboot(name, call=None):
    """
    reboot a machine by name
    :param name: name given to the machine
    :param call: call value in this case is 'action'
    :return: true if successful

    CLI Example:

    .. code-block:: bash

        salt-cloud -a reboot vm_name
    """
    datacenter_id = get_datacenter_id()
    conn = get_conn()
    node = get_node(conn, name)

    conn.reboot_server(datacenter_id=datacenter_id, server_id=node["id"])

    return True


def stop(name, call=None):
    """
    stop a machine by name
    :param name: name given to the machine
    :param call: call value in this case is 'action'
    :return: true if successful

    CLI Example:

    .. code-block:: bash

        salt-cloud -a stop vm_name
    """
    datacenter_id = get_datacenter_id()
    conn = get_conn()
    node = get_node(conn, name)

    conn.stop_server(datacenter_id=datacenter_id, server_id=node["id"])

    return True


def start(name, call=None):
    """
    start a machine by name
    :param name: name given to the machine
    :param call: call value in this case is 'action'
    :return: true if successful

    CLI Example:

    .. code-block:: bash

        salt-cloud -a start vm_name
    """
    datacenter_id = get_datacenter_id()
    conn = get_conn()
    node = get_node(conn, name)

    conn.start_server(datacenter_id=datacenter_id, server_id=node["id"])

    return True


def _override_size(vm_):
    """
    Apply any extra component overrides to VM from the cloud profile.
    """
    vm_size = get_size(vm_)

    if "cores" in vm_:
        vm_size["cores"] = vm_["cores"]

    if "ram" in vm_:
        vm_size["ram"] = vm_["ram"]

    return vm_size


def _get_server(vm_, volumes, nics):
    """
    Construct server instance from cloud profile config
    """
    # Apply component overrides to the size from the cloud profile config
    vm_size = _override_size(vm_)

    # Set the server availability zone from the cloud profile config
    availability_zone = config.get_cloud_config_value(
        "availability_zone", vm_, __opts__, default=None, search_global=False
    )

    # Assign CPU family from the cloud profile config
    cpu_family = config.get_cloud_config_value(
        "cpu_family", vm_, __opts__, default=None, search_global=False
    )

    # Contruct server object
    return Server(
        name=vm_["name"],
        ram=vm_size["ram"],
        availability_zone=availability_zone,
        cores=vm_size["cores"],
        cpu_family=cpu_family,
        create_volumes=volumes,
        nics=nics,
    )


def _get_system_volume(vm_):
    """
    Construct VM system volume list from cloud profile config
    """

    # Override system volume size if 'disk_size' is defined in cloud profile
    disk_size = get_size(vm_)["disk"]
    if "disk_size" in vm_:
        disk_size = vm_["disk_size"]

    # Construct the system volume
    volume = Volume(
        name="{} Storage".format(vm_["name"]),
        size=disk_size,
        disk_type=get_disk_type(vm_),
    )

    if "image_password" in vm_:
        image_password = vm_["image_password"]
        volume.image_password = image_password

    # Retrieve list of SSH public keys
    ssh_keys = get_public_keys(vm_)
    volume.ssh_keys = ssh_keys

    if "image_alias" in vm_.keys():
        volume.image_alias = vm_["image_alias"]
    else:
        volume.image = get_image(vm_)["id"]
        # Set volume availability zone if defined in the cloud profile
        if "disk_availability_zone" in vm_:
            volume.availability_zone = vm_["disk_availability_zone"]

    return volume


def _get_data_volumes(vm_):
    """
    Construct a list of optional data volumes from the cloud profile
    """
    ret = []
    volumes = vm_["volumes"]
    for key, value in volumes.items():
        # Verify the required 'disk_size' property is present in the cloud
        # profile config
        if "disk_size" not in volumes[key].keys():
            raise SaltCloudConfigError(f"The volume '{key}' is missing 'disk_size'")
        # Use 'HDD' if no 'disk_type' property is present in cloud profile
        if "disk_type" not in volumes[key].keys():
            volumes[key]["disk_type"] = "HDD"

        # Construct volume object and assign to a list.
        volume = Volume(
            name=key,
            size=volumes[key]["disk_size"],
            disk_type=volumes[key]["disk_type"],
            licence_type="OTHER",
        )

        # Set volume availability zone if defined in the cloud profile
        if "disk_availability_zone" in volumes[key].keys():
            volume.availability_zone = volumes[key]["disk_availability_zone"]

        ret.append(volume)

    return ret


def _get_ip_addresses(ip_addresses):
    """
    Construct a list of ip address
    """
    ret = []
    for item in ip_addresses:
        ret.append(item)

    return ret


def _get_firewall_rules(firewall_rules):
    """
    Construct a list of optional firewall rules from the cloud profile.
    """
    ret = []
    for key, value in firewall_rules.items():
        # Verify the required 'protocol' property is present in the cloud
        # profile config
        if "protocol" not in firewall_rules[key].keys():
            raise SaltCloudConfigError(
                f"The firewall rule '{key}' is missing 'protocol'"
            )
        ret.append(
            FirewallRule(
                name=key,
                protocol=firewall_rules[key].get("protocol", None),
                source_mac=firewall_rules[key].get("source_mac", None),
                source_ip=firewall_rules[key].get("source_ip", None),
                target_ip=firewall_rules[key].get("target_ip", None),
                port_range_start=firewall_rules[key].get("port_range_start", None),
                port_range_end=firewall_rules[key].get("port_range_end", None),
                icmp_type=firewall_rules[key].get("icmp_type", None),
                icmp_code=firewall_rules[key].get("icmp_code", None),
            )
        )

    return ret


def _wait_for_completion(conn, promise, wait_timeout, msg):
    """
    Poll request status until resource is provisioned.
    """
    if not promise:
        return
    wait_timeout = time.time() + wait_timeout
    while wait_timeout > time.time():
        time.sleep(5)
        operation_result = conn.get_request(
            request_id=promise["requestId"], status=True
        )

        if operation_result["metadata"]["status"] == "DONE":
            return
        elif operation_result["metadata"]["status"] == "FAILED":
            raise Exception(
                "Request: {}, requestId: {} failed to complete:\n{}".format(
                    msg,
                    str(promise["requestId"]),
                    operation_result["metadata"]["message"],
                )
            )

    raise Exception(
        'Timed out waiting for asynchronous operation {} "{}" to complete.'.format(
            msg, str(promise["requestId"])
        )
    )



"""
Connection module for Amazon CloudWatch

.. versionadded:: 2014.7.0

:configuration: This module accepts explicit credentials but can also utilize
    IAM roles assigned to the instance through Instance Profiles. Dynamic
    credentials are then automatically obtained from AWS API and no further
    configuration is necessary. More Information available at:

    .. code-block:: text

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file:

    .. code-block:: yaml

        cloudwatch.keyid: GKTADJGHEIQSXMKKRBJ08H
        cloudwatch.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration:

    .. code-block:: yaml

        cloudwatch.region: us-east-1

    If a region is not specified, the default is us-east-1.

    It's also possible to specify key, keyid and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

:depends: boto
"""

# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602

import logging

import yaml  # pylint: disable=blacklisted-import

import salt.utils.json
import salt.utils.odict as odict
import salt.utils.versions

try:
    import boto
    import boto.ec2.cloudwatch
    import boto.ec2.cloudwatch.dimension
    import boto.ec2.cloudwatch.listelement

    logging.getLogger("boto").setLevel(logging.CRITICAL)
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if boto libraries exist.
    """
    has_boto_reqs = salt.utils.versions.check_boto_reqs(check_boto3=False)
    if has_boto_reqs is True:
        __utils__["boto.assign_funcs"](
            __name__, "cloudwatch", module="ec2.cloudwatch", pack=__salt__
        )
    return has_boto_reqs


def get_alarm(name, region=None, key=None, keyid=None, profile=None):
    """
    Get alarm details. Also can be used to check to see if an alarm exists.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudwatch.get_alarm myalarm region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    alarms = conn.describe_alarms(alarm_names=[name])
    if not alarms:
        return None
    if len(alarms) > 1:
        log.error("multiple alarms matched name '%s'", name)
    return _metric_alarm_to_dict(alarms[0])


def _safe_dump(data):
    """
    this presenter magic makes yaml.safe_dump
    work with the objects returned from
    boto.describe_alarms()
    """
    custom_dumper = __utils__["yaml.get_dumper"]("SafeOrderedDumper")

    def boto_listelement_presenter(dumper, data):
        return dumper.represent_list(list(data))

    yaml.add_representer(
        boto.ec2.cloudwatch.listelement.ListElement,
        boto_listelement_presenter,
        Dumper=custom_dumper,
    )

    def dimension_presenter(dumper, data):
        return dumper.represent_dict(dict(data))

    yaml.add_representer(
        boto.ec2.cloudwatch.dimension.Dimension,
        dimension_presenter,
        Dumper=custom_dumper,
    )

    return __utils__["yaml.dump"](data, Dumper=custom_dumper)


def get_all_alarms(region=None, prefix=None, key=None, keyid=None, profile=None):
    """
    Get all alarm details.  Produces results that can be used to create an sls
    file.

    If prefix parameter is given, alarm names in the output will be prepended
    with the prefix; alarms that have the prefix will be skipped.  This can be
    used to convert existing alarms to be managed by salt, as follows:

        1. Make a "backup" of all existing alarms
            $ salt-call boto_cloudwatch.get_all_alarms --out=txt | sed "s/local: //" > legacy_alarms.sls

        2. Get all alarms with new prefixed names
            $ salt-call boto_cloudwatch.get_all_alarms "prefix=**MANAGED BY SALT** " --out=txt | sed "s/local: //" > managed_alarms.sls

        3. Insert the managed alarms into cloudwatch
            $ salt-call state.template managed_alarms.sls

        4.  Manually verify that the new alarms look right

        5.  Delete the original alarms
            $ sed s/present/absent/ legacy_alarms.sls > remove_legacy_alarms.sls
            $ salt-call state.template remove_legacy_alarms.sls

        6.  Get all alarms again, verify no changes
            $ salt-call boto_cloudwatch.get_all_alarms --out=txt | sed "s/local: //" > final_alarms.sls
            $ diff final_alarms.sls managed_alarms.sls

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudwatch.get_all_alarms region=us-east-1 --out=txt
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    alarms = conn.describe_alarms()
    results = odict.OrderedDict()
    for alarm in alarms:
        alarm = _metric_alarm_to_dict(alarm)
        name = alarm["name"]
        if prefix:
            if name.startswith(prefix):
                continue
            name = prefix + alarm["name"]
        del alarm["name"]
        alarm_sls = [{"name": name}, {"attributes": alarm}]
        results["manage alarm " + name] = {"boto_cloudwatch_alarm.present": alarm_sls}
    return _safe_dump(results)


def create_or_update_alarm(
    connection=None,
    name=None,
    metric=None,
    namespace=None,
    statistic=None,
    comparison=None,
    threshold=None,
    period=None,
    evaluation_periods=None,
    unit=None,
    description="",
    dimensions=None,
    alarm_actions=None,
    insufficient_data_actions=None,
    ok_actions=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Create or update a cloudwatch alarm.

    Params are the same as:
        https://boto.readthedocs.io/en/latest/ref/cloudwatch.html#boto.ec2.cloudwatch.alarm.MetricAlarm.

    Dimensions must be a dict. If the value of Dimensions is a string, it will
    be json decoded to produce a dict. alarm_actions, insufficient_data_actions,
    and ok_actions must be lists of string.  If the passed-in value is a string,
    it will be split on "," to produce a list. The strings themselves for
    alarm_actions, insufficient_data_actions, and ok_actions must be Amazon
    resource names (ARN's); however, this method also supports an arn lookup
    notation, as follows:

        arn:aws:....                                    ARN as per http://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html
        scaling_policy:<as_name>:<scaling_policy_name>  The named autoscale group scaling policy, for the named group (e.g.  scaling_policy:my-asg:ScaleDown)

    This is convenient for setting up autoscaling as follows.  First specify a
    boto_asg.present state for an ASG with scaling_policies, and then set up
    boto_cloudwatch_alarm.present states which have alarm_actions that
    reference the scaling_policy.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudwatch.create_alarm name=myalarm ... region=us-east-1
    """
    # clean up argument types, so that CLI works
    if threshold:
        threshold = float(threshold)
    if period:
        period = int(period)
    if evaluation_periods:
        evaluation_periods = int(evaluation_periods)
    if isinstance(dimensions, str):
        dimensions = salt.utils.json.loads(dimensions)
        if not isinstance(dimensions, dict):
            log.error(
                "could not parse dimensions argument: must be json encoding of a dict:"
                " '%s'",
                dimensions,
            )
            return False
    if isinstance(alarm_actions, str):
        alarm_actions = alarm_actions.split(",")
    if isinstance(insufficient_data_actions, str):
        insufficient_data_actions = insufficient_data_actions.split(",")
    if isinstance(ok_actions, str):
        ok_actions = ok_actions.split(",")

    # convert provided action names into ARN's
    if alarm_actions:
        alarm_actions = convert_to_arn(
            alarm_actions, region=region, key=key, keyid=keyid, profile=profile
        )
    if insufficient_data_actions:
        insufficient_data_actions = convert_to_arn(
            insufficient_data_actions,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )
    if ok_actions:
        ok_actions = convert_to_arn(
            ok_actions, region=region, key=key, keyid=keyid, profile=profile
        )

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    alarm = boto.ec2.cloudwatch.alarm.MetricAlarm(
        connection=connection,
        name=name,
        metric=metric,
        namespace=namespace,
        statistic=statistic,
        comparison=comparison,
        threshold=threshold,
        period=period,
        evaluation_periods=evaluation_periods,
        unit=unit,
        description=description,
        dimensions=dimensions,
        alarm_actions=alarm_actions,
        insufficient_data_actions=insufficient_data_actions,
        ok_actions=ok_actions,
    )
    conn.create_alarm(alarm)
    log.info("Created/updated alarm %s", name)
    return True


def convert_to_arn(arns, region=None, key=None, keyid=None, profile=None):
    """
    Convert a list of strings into actual arns. Converts convenience names such
    as 'scaling_policy:...'

    CLI Example:

    .. code-block:: bash

        salt '*' convert_to_arn 'scaling_policy:'
    """
    results = []
    for arn in arns:
        if arn.startswith("scaling_policy:"):
            _, as_group, scaling_policy_name = arn.split(":")
            policy_arn = __salt__["boto_asg.get_scaling_policy_arn"](
                as_group, scaling_policy_name, region, key, keyid, profile
            )
            if policy_arn:
                results.append(policy_arn)
            else:
                log.error("Could not convert: %s", arn)
        else:
            results.append(arn)
    return results


def delete_alarm(name, region=None, key=None, keyid=None, profile=None):
    """
    Delete a cloudwatch alarm

    CLI example to delete a queue::

        salt myminion boto_cloudwatch.delete_alarm myalarm region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    conn.delete_alarms([name])
    log.info("Deleted alarm %s", name)
    return True


def _metric_alarm_to_dict(alarm):
    """
    Convert a boto.ec2.cloudwatch.alarm.MetricAlarm into a dict. Convenience
    for pretty printing.
    """
    d = odict.OrderedDict()
    fields = [
        "name",
        "metric",
        "namespace",
        "statistic",
        "comparison",
        "threshold",
        "period",
        "evaluation_periods",
        "unit",
        "description",
        "dimensions",
        "alarm_actions",
        "insufficient_data_actions",
        "ok_actions",
    ]
    for f in fields:
        if hasattr(alarm, f):
            d[f] = getattr(alarm, f)
    return d



"""
Using states instead of maps to deploy clouds
=============================================

.. versionadded:: 2014.1.0

Use this minion to spin up a cloud instance:

.. code-block:: yaml

    my-ec2-instance:
      cloud.profile:
        my-ec2-config
"""

import pprint

import salt.utils.cloud as suc


def __virtual__():
    """
    Only load if the cloud module is available in __salt__
    """
    if "cloud.profile" in __salt__:
        return True
    return (False, "cloud module could not be loaded")


def _check_name(name):
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}
    if suc.check_name(name, "a-zA-Z0-9._-"):
        ret["comment"] = "Invalid characters in name."
        ret["result"] = False
        return ret
    else:
        ret["result"] = True
        return ret


def _valid(name, comment="", changes=None):
    if not changes:
        changes = {}
    return {"name": name, "result": True, "changes": changes, "comment": comment}


def _get_instance(names):
    # for some reason loader overwrites __opts__['test'] with default
    # value of False, thus store and then load it again after action
    test = __opts__.get("test", False)
    instance = __salt__["cloud.action"](fun="show_instance", names=names)
    __opts__["test"] = test
    return instance


def present(name, cloud_provider, onlyif=None, unless=None, opts=None, **kwargs):
    """
    Spin up a single instance on a cloud provider, using salt-cloud. This state
    does not take a profile argument; rather, it takes the arguments that would
    normally be configured as part of the state.

    Note that while this function does take any configuration argument that
    would normally be used to create an instance, it will not verify the state
    of any of those arguments on an existing instance. Stateful properties of
    an instance should be configured using their own individual state (i.e.,
    cloud.tagged, cloud.untagged, etc).

    name
        The name of the instance to create

    cloud_provider
        The name of the cloud provider to use

    onlyif
        Do run the state only if is unless succeed

    unless
        Do not run the state at least unless succeed

    opts
        Any extra opts that need to be used
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    retcode = __salt__["cmd.retcode"]
    if onlyif is not None:
        if not isinstance(onlyif, str):
            if not onlyif:
                return _valid(name, comment="onlyif condition is false")
        elif isinstance(onlyif, str):
            if retcode(onlyif, python_shell=True) != 0:
                return _valid(name, comment="onlyif condition is false")
    if unless is not None:
        if not isinstance(unless, str):
            if unless:
                return _valid(name, comment="unless condition is true")
        elif isinstance(unless, str):
            if retcode(unless, python_shell=True) == 0:
                return _valid(name, comment="unless condition is true")

    # provider=None not cloud_provider because
    # need to ensure ALL providers don't have the instance
    if __salt__["cloud.has_instance"](name=name, provider=None):
        ret["result"] = True
        ret["comment"] = f"Already present instance {name}"
        return ret

    if __opts__["test"]:
        ret["comment"] = f"Instance {name} needs to be created"
        return ret

    info = __salt__["cloud.create"](cloud_provider, name, opts=opts, **kwargs)
    if info and "Error" not in info:
        ret["changes"] = info
        ret["result"] = True
        ret["comment"] = (
            "Created instance {} using provider {} and the following options: {}".format(
                name, cloud_provider, pprint.pformat(kwargs)
            )
        )
    elif info and "Error" in info:
        ret["result"] = False
        ret["comment"] = "Failed to create instance {} using profile {}: {}".format(
            name,
            profile,
            info["Error"],
        )
    else:
        ret["result"] = False
        ret["comment"] = (
            "Failed to create instance {} using profile {}, "
            "please check your configuration".format(name, profile)
        )
    return ret


def absent(name, onlyif=None, unless=None):
    """
    Ensure that no instances with the specified names exist.

    CAUTION: This is a destructive state, which will search all
    configured cloud providers for the named instance,
    and destroy it.

    name
        The name of the instance to destroy

    onlyif
        Do run the state only if is unless succeed

    unless
        Do not run the state at least unless succeed

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}
    retcode = __salt__["cmd.retcode"]

    if onlyif is not None:
        if not isinstance(onlyif, str):
            if not onlyif:
                return _valid(name, comment="onlyif condition is false")
        elif isinstance(onlyif, str):
            if retcode(onlyif, python_shell=True) != 0:
                return _valid(name, comment="onlyif condition is false")
    if unless is not None:
        if not isinstance(unless, str):
            if unless:
                return _valid(name, comment="unless condition is true")
        elif isinstance(unless, str):
            if retcode(unless, python_shell=True) == 0:
                return _valid(name, comment="unless condition is true")

    if not __salt__["cloud.has_instance"](name=name, provider=None):
        ret["result"] = True
        ret["comment"] = f"Already absent instance {name}"
        return ret

    if __opts__["test"]:
        ret["comment"] = f"Instance {name} needs to be destroyed"
        return ret

    info = __salt__["cloud.destroy"](name)
    if info and "Error" not in info:
        ret["changes"] = info
        ret["result"] = True
        ret["comment"] = f"Destroyed instance {name}"
    elif "Error" in info:
        ret["result"] = False
        ret["comment"] = "Failed to destroy instance {}: {}".format(
            name,
            info["Error"],
        )
    else:
        ret["result"] = False
        ret["comment"] = f"Failed to destroy instance {name}"
    return ret


def profile(name, profile, onlyif=None, unless=None, opts=None, **kwargs):
    """
    Create a single instance on a cloud provider, using a salt-cloud profile.

    Note that while profiles used this function do take any configuration
    argument that would normally be used to create an instance using a profile,
    this state will not verify the state of any of those arguments on an
    existing instance. Stateful properties of an instance should be configured
    using their own individual state (i.e., cloud.tagged, cloud.untagged, etc).

    name
        The name of the instance to create

    profile
        The name of the cloud profile to use

    onlyif
        Do run the state only if is unless succeed

    unless
        Do not run the state at least unless succeed

    kwargs
        Any profile override or addition

    opts
        Any extra opts that need to be used
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}
    retcode = __salt__["cmd.retcode"]
    if onlyif is not None:
        if not isinstance(onlyif, str):
            if not onlyif:
                return _valid(name, comment="onlyif condition is false")
        elif isinstance(onlyif, str):
            if retcode(onlyif, python_shell=True) != 0:
                return _valid(name, comment="onlyif condition is false")
    if unless is not None:
        if not isinstance(unless, str):
            if unless:
                return _valid(name, comment="unless condition is true")
        elif isinstance(unless, str):
            if retcode(unless, python_shell=True) == 0:
                return _valid(name, comment="unless condition is true")
    instance = _get_instance([name])
    if instance and not any("Not Actioned" in key for key in instance):
        ret["result"] = True
        ret["comment"] = f"Already present instance {name}"
        return ret

    if __opts__["test"]:
        ret["comment"] = f"Instance {name} needs to be created"
        return ret

    info = __salt__["cloud.profile"](profile, name, vm_overrides=kwargs, opts=opts)

    # get either {Error: ''} or {namestring: {Error: ''}}
    # which is what we can get from providers returns
    main_error = info.get("Error", "")
    name_error = ""
    if isinstance(info, dict):
        subinfo = info.get(name, {})
        if isinstance(subinfo, dict):
            name_error = subinfo.get("Error", None)
    error = main_error or name_error
    if info and not error:
        node_info = info.get(name)
        ret["result"] = True
        default_msg = "Created instance {} using profile {}".format(
            name,
            profile,
        )
        # some providers support changes
        if "changes" in node_info:
            ret["changes"] = node_info["changes"]
            ret["comment"] = node_info.get("comment", default_msg)
        else:
            ret["changes"] = info
            ret["comment"] = default_msg
    elif error:
        ret["result"] = False
        ret["comment"] = "Failed to create instance {} using profile {}: {}".format(
            name,
            profile,
            f"{main_error}\n{name_error}\n".strip(),
        )
    else:
        ret["result"] = False
        ret["comment"] = "Failed to create instance {} using profile {}".format(
            name,
            profile,
        )
    return ret


def volume_present(name, provider=None, **kwargs):
    """
    Check that a block volume exists.
    """
    ret = _check_name(name)
    if not ret["result"]:
        return ret

    volumes = __salt__["cloud.volume_list"](provider=provider)

    if name in volumes:
        ret["comment"] = f"Volume exists: {name}"
        ret["result"] = True
        return ret
    elif __opts__["test"]:
        ret["comment"] = f"Volume {name} will be created."
        ret["result"] = None
        return ret

    response = __salt__["cloud.volume_create"](names=name, provider=provider, **kwargs)
    if response:
        ret["result"] = True
        ret["comment"] = f"Volume {name} was created"
        ret["changes"] = {"old": None, "new": response}
    else:
        ret["result"] = False
        ret["comment"] = f"Volume {name} failed to create."
    return ret


def volume_absent(name, provider=None, **kwargs):
    """
    Check that a block volume exists.
    """
    ret = _check_name(name)
    if not ret["result"]:
        return ret

    volumes = __salt__["cloud.volume_list"](provider=provider)

    if name not in volumes:
        ret["comment"] = "Volume is absent."
        ret["result"] = True
        return ret
    elif __opts__["test"]:
        ret["comment"] = f"Volume {name} will be deleted."
        ret["result"] = None
        return ret

    response = __salt__["cloud.volume_delete"](names=name, provider=provider, **kwargs)
    if response:
        ret["result"] = True
        ret["comment"] = f"Volume {name} was deleted"
        ret["changes"] = {"old": volumes[name], "new": response}
    else:
        ret["result"] = False
        ret["comment"] = f"Volume {name} failed to delete."
    return ret


def volume_attached(name, server_name, provider=None, **kwargs):
    """
    Check if a block volume is attached.
    """
    ret = _check_name(name)
    if not ret["result"]:
        return ret

    ret = _check_name(server_name)
    if not ret["result"]:
        return ret

    volumes = __salt__["cloud.volume_list"](provider=provider)
    instance = __salt__["cloud.action"](fun="show_instance", names=server_name)

    if name in volumes and volumes[name]["attachments"]:
        volume = volumes[name]
        ret["comment"] = "Volume {name} is already attached: {attachments}".format(
            **volumes[name]
        )
        ret["result"] = True
        return ret
    elif name not in volumes:
        ret["comment"] = f"Volume {name} does not exist"
        ret["result"] = False
        return ret
    elif not instance:
        ret["comment"] = f"Server {server_name} does not exist"
        ret["result"] = False
        return ret
    elif __opts__["test"]:
        ret["comment"] = f"Volume {name} will be will be attached."
        ret["result"] = None
        return ret

    response = __salt__["cloud.volume_attach"](
        provider=provider, names=name, server_name=server_name, **kwargs
    )
    if response:
        ret["result"] = True
        ret["comment"] = f"Volume {name} was created"
        ret["changes"] = {"old": volumes[name], "new": response}
    else:
        ret["result"] = False
        ret["comment"] = f"Volume {name} failed to attach."
    return ret


def volume_detached(name, server_name=None, provider=None, **kwargs):
    """
    Check if a block volume is attached.

    Returns True if server or Volume do not exist.
    """
    ret = _check_name(name)
    if not ret["result"]:
        return ret

    if server_name is not None:
        ret = _check_name(server_name)
        if not ret["result"]:
            return ret

    volumes = __salt__["cloud.volume_list"](provider=provider)
    if server_name:
        instance = __salt__["cloud.action"](fun="show_instance", names=[name])
    else:
        instance = None

    if name in volumes and not volumes[name]["attachments"]:
        volume = volumes[name]
        ret["comment"] = "Volume {name} is not currently attached to anything.".format(
            **volumes[name]
        )
        ret["result"] = True
        return ret
    elif name not in volumes:
        ret["comment"] = f"Volume {name} does not exist"
        ret["result"] = True
        return ret
    elif not instance and server_name is not None:
        ret["comment"] = f"Server {server_name} does not exist"
        ret["result"] = True
        return ret
    elif __opts__["test"]:
        ret["comment"] = f"Volume {name} will be will be detached."
        ret["result"] = None
        return ret

    response = __salt__["cloud.volume_detach"](
        provider=provider, names=name, server_name=server_name, **kwargs
    )
    if response:
        ret["result"] = True
        ret["comment"] = f"Volume {name} was created"
        ret["changes"] = {"old": volumes[name], "new": response}
    else:
        ret["result"] = False
        ret["comment"] = f"Volume {name} failed to detach."
    return ret
